# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to obtain and display session information.

import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        cur.execute ('help session')
        row = cur.fetchone ()
        for i in range (0, len (row)):
            print ('{:>40}  {}'.format (cur.description [i][0], row [i]))



import re
import ntpath
import os.path
import errno
import pandas as pd
from snowfinch.logger import log
from snowfinch.exceptions import *


logger = log.get_logger()


def get_object_metadata(conn, object_name: str, object_type: str, object_schema: str):
    if object_type == 'table':
        metadata_sql = f"""
        SELECT table_catalog, table_schema, table_owner, table_type, row_count,last_altered
        FROM information_schema.tables 
        WHERE  
            upper(table_schema) = '{object_schema}'
            AND upper(table_name) = '{object_name}';
        """
    elif object_type == 'procedure':
        metadata_sql = f"""
        SELECT procedure_catalog, procedure_schema, procedure_owner, last_altered
        FROM information_schema.procedures 
        WHERE  
            upper(procedure_schema) = '{object_schema}'
            AND upper(procedure_name) = '{object_name}';
        """
    else:
        metadata_sql = ""
        print("Object type currently not part of this method.")

    print(f"\nsql query for {object_schema}.{object_name}")
    print(metadata_sql)
    try:
        print("\nFetching Metadata  from SnowFlake...")
        # object_metadata = conn.execute(metadata_sql).fetchall()
        object_metadata = pd.read_sql(metadata_sql, conn)
        print(object_metadata.transpose())

    except ProgrammingError as err:
        print("ProgrammingError", err)
        raise

    return object_metadata


def get_object_owner(conn, object_name, object_type, object_schema):
    if object_type == 'table':
        owner_query = f"""
        SELECT table_owner as owner
        FROM information_schema.tables
        WHERE
        upper(table_schema) = '{object_schema}'
        AND upper(table_name) = '{object_name}'
        """
    elif object_type == 'procedure':
        owner_query = f"""
        SELECT procedure_owner as owner
        FROM information_schema.procedure
        WHERE
        upper(table_schema) = '{object_schema}'
        AND upper(procedure_name) = '{object_name}'
        """
    else:
        owner_query = ""
        logger.info("Object type currently not part of this method.")

    print(f"\nsql query for {object_schema}.{object_name}")
    print(owner_query)

    try:
        logger.info("\nFetching table owner Metadata from SnowFlake...")
        df = pd.read_sql(owner_query, conn)
        object_owner = ''.join(df['owner'])
        print(f"object_owner: {object_owner}")
    except ProgrammingError as err:
        logger.info("ProgrammingError", err)
        raise
    return object_owner


def grant_ownership(conn, object_name, object_type, object_schema, object_owner):
    if object_type == 'table':
        grant_cmd = f"""
        grant ownership on table {object_schema}.{object_name} to
               role {object_owner} copy current grants;
        """
    elif object_type == 'procedure':
        grant_cmd = f"""
        grant ownership on procedure {object_schema}.{object_name} to
               role {object_owner} copy current grants;
        """
    else:
        grant_cmd = ""
        logger.info("Object type currently not part of this method.")

    logger.info(grant_cmd)
    try:
        print(f"\nChanging the ownership to {object_owner}...")
        conn.execute(grant_cmd)

    except ProgrammingError as err:
        logger.info("ProgrammingError", err)
        raise


def is_word_in_string(word, text):
    pattern = r'(^|[^\w]){}([^\w]|$)'.format(word)
    pattern = re.compile(pattern, re.IGNORECASE)
    matches = re.search(pattern, text)
    return bool(matches)


def read_sql(sql, conn):
    try:
        # read data using a sql query
        _df = pd.read_sql_query(sql, conn)

        # trim columns
        _df = _df.applymap(lambda x: x.strip() if isinstance(x, str) else x)

        # upper case columns
        _df.columns = [x.upper() for x in _df.columns]
        return _df

    except OperationalError as err:
        logger.Info('Error occurred while executing a query {}'.format(err.args))


# from each line, remove all text which is behind a '--'
def cut_comment(query: str) -> str:
    idx = query.find('--')
    if idx >= 0:
        query = query[:idx]
    return query


def get_sql_stmt_list(file_path, delimiter=';'):
    print(delimiter)
    """
    this function reads one or more SQL statements from a script and returns
    them in a list.

    :param file_path: path to file that contains SQL text to be executed by Snowflake.
    :param delimiter: the delimiter used to separate independent SQL statements.
    :throws: snowflake.connector.errors.ProgrammingError
    :returns: sql statement
    :rtype: str
    """
    with open(file_path, 'r', encoding='utf-8') as file:
        queries = [line.strip() for line in file.readlines()]
        queries = [cut_comment(q) for q in queries]
        sql_command = ' '.join(queries)
        return sql_command


def get_outfile_full_name(_infile, extension):
    filename = ntpath.basename(_infile)
    filepath = ntpath.dirname(_infile)
    out_file_name = filename.split(".")[0]
    out_file = f"{filepath}/{out_file_name}.{extension}"
    return out_file


def silent_remove(file):
    try:
        os.remove(file)
    except OSError as e:
        if e.errno != errno.ENOENT:  # no such file or directory
            raise  # re-raise exception if a different error occurred


def execute_sql_from_file(filename, conn):
    # Execute every command from the input file
    with open(filename, 'r') as fd:
        for command in fd.read().split(';'):
            # This will skip and report errors
            # For example, if the tables do not yet exist, this will skip over
            # the DROP TABLE commands
            try:
                conn.execute(command)
            except OperationalError as err:
                logger.error("Command skipped: ", err)



# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to use the commit and rollback methods after turning off auto-commit.

import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        cur.execute ("{fn teradata_nativesql}{fn teradata_autocommit_off}")

        cur.execute ("create volatile table voltab (c1 integer) on commit preserve rows")
        con.commit ()

        cur.execute ("insert into voltab values (1)")
        con.commit ()

        cur.execute ("insert into voltab values (2)")
        con.rollback ()

        cur.execute ("insert into voltab values (3)")
        cur.execute ("insert into voltab values (4)")
        con.commit ()

        cur.execute ("insert into voltab values (5)")
        cur.execute ("insert into voltab values (6)")
        con.rollback ()

        cur.execute ("select * from voltab order by 1")
        con.commit ()

        anValues = [ row [0] for row in cur.fetchall () ]

        print ("Expected result set row values: [1, 3, 4]")
        print ("Obtained result set row values: {}".format (anValues))



# Copyright 2021 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to display the output from a show command.
# The show command output includes embedded carriage returns as line terminators,
# which typically must be changed to newlines in order to display properly.
# The show command output may include multiple result sets, and the cursor nextset
# function must be called to advance to subsequent result sets.

import teradatasql

def ShowCommand (cur, s):
    print ("-- " + s)
    cur.execute (s)
    n = 1
    while True:
        print ("-- result set {}".format (n))
        rows = cur.fetchall ()
        for row in rows:
            print ("".join ([ str (col).replace ("\r", "\n") for col in row ]))
        if cur.nextset ():
            n += 1
        else:
            break

with teradatasql.connect (host="whomooz", user="guest", password="please") as con:
    with con.cursor () as cur:
        ShowCommand (cur, "show view DBC.DBCInfo")
        print ()
        ShowCommand (cur, "show select * from DBC.DBCInfo")



# Copyright 2020 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to specify the data type of parameter marker bind values.
# Use the escape function teradata_parameter to override the data type for a parameter marker bind value.

import teradatasql

with teradatasql.connect (host="whomooz", user="guest", password="please") as con:
    with con.cursor () as cur:
        aao = [
            [ "BYTEINT"                          , 1 ],
            [ "SMALLINT"                         , 2 ],
            [ "INTEGER"                          , 3 ],
            [ "DECIMAL(2)"                       , "4" ],
            [ "DECIMAL(4)"                       , "5" ],
            [ "DECIMAL(9)"                       , "6" ],
            [ "DECIMAL(18)"                      , "7" ],
            [ "DECIMAL(38)"                      , "8" ],
            [ "PERIOD(DATE)"                     , "0009-09-09,0909-09-09" ],
            [ "PERIOD(TIME)"                     , "10:10:10.101000,10:10:10.101010" ],
            [ "PERIOD(TIME WITH TIME ZONE)"      , "11:11:11.111110+11:11,11:11:11.111111+11:11" ],
            [ "PERIOD(TIMESTAMP)"                , "1212-01-12 12:12:12.121212,1212-02-12 12:12:12.121212" ],
            [ "PERIOD(TIMESTAMP WITH TIME ZONE)" , "1313-01-13 13:13:13.131313+13:13,1313-02-13 13:13:13.131313+13:13" ],
            [ "INTERVAL YEAR(4)"                 , "-1414" ],
            [ "INTERVAL YEAR(4) TO MONTH"        , "-1515-11" ],
            [ "INTERVAL MONTH(4)"                , "-1616" ],
            [ "INTERVAL DAY(4)"                  , "-1717" ],
            [ "INTERVAL DAY(4) TO HOUR"          , "-1818 18" ],
            [ "INTERVAL DAY(4) TO MINUTE"        , "-1919 19:19" ],
            [ "INTERVAL DAY(4) TO SECOND"        , "-2020 20:20:20.202020" ],
            [ "INTERVAL HOUR(4)"                 , "-2121" ],
            [ "INTERVAL HOUR(4) TO MINUTE"       , "-2222:22" ],
            [ "INTERVAL HOUR(4) TO SECOND"       , "-2323:23:23.232323" ],
            [ "INTERVAL MINUTE(4)"               , "-2424" ],
            [ "INTERVAL MINUTE(4) TO SECOND"     , "-2525:25.252525" ],
            [ "INTERVAL SECOND(4)"               , "-2626.262626" ],
            [ "BYTE(2)"                          , b'27' ],
            [ "CHAR(2)"                          , "28" ],
        ]

        print ("   teradata_parameter bind type      Result from TYPE function            Round-trip data value")
        print ("   --------------------------------  -----------------------------------  ---------------------")

        for n, ao in enumerate (aao, 1):
            sDataType, oValue = ao [0], ao [1]

            # The result from the built-in TYPE function shows the data type that the database receives from the driver.
            cur.execute ("{fn teradata_parameter(1," + sDataType + ")}select trim(type(?))", [ None ])
            sTypeFuncResult = cur.fetchone () [0]

            # The database does not support binding a PERIOD value to a select-list parameter marker, so we go through a table.
            if sDataType.startswith ("PERIOD"):
                cur.execute ("create volatile table volatiletable (c1 " + sDataType + ") no primary index on commit preserve rows")
                cur.execute ("{fn teradata_parameter(1," + sDataType + ")}insert into volatiletable values (?)", [ oValue ])
                cur.execute ("select * from volatiletable")
                oOutput = cur.fetchone () [0]
                cur.execute ("drop table volatiletable")

            else: # Other data types can be bound to a select-list parameter marker.
                cur.execute ("{fn teradata_parameter(1," + sDataType + ")}select ?", [ oValue ])
                oOutput = cur.fetchone () [0]

            print ("{:>2} {:<33} {:<36} {}".format (n, sDataType, sTypeFuncResult, oOutput))



# Copyright 2019 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to work with the Teradata Database's Character Export Width behavior.

# The Teradata SQL Driver for Python always uses the UTF8 session character set, and the charset connection
# parameter is not supported. Be aware of the Teradata Database's Character Export Width behavior that adds trailing
# space padding to fixed-width CHAR data type result set column values when using the UTF8 session character set.

# Work around this drawback by using CAST or TRIM in SQL SELECT statements, or in views, to convert fixed-width CHAR
# data types to VARCHAR.

import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        cur.execute ("CREATE TABLE MyTable (c1 CHAR(10), c2 CHAR(10))")
        try:
            cur.execute ("INSERT INTO MyTable VALUES ('a', 'b')")

            print ("Original query that produces trailing space padding:")
            cur.execute ("SELECT c1, c2 FROM MyTable")
            [ print (row) for row in cur.fetchall () ]

            print ("Modified query with either CAST or TRIM to avoid trailing space padding:")
            cur.execute ("SELECT CAST(c1 AS VARCHAR(10)), TRIM(TRAILING FROM c1) FROM MyTable")
            [ print (row) for row in cur.fetchall () ]

            cur.execute ("CREATE VIEW MyView (c1, c2) AS SELECT CAST(c1 AS VARCHAR(10)), TRIM(TRAILING FROM c2) FROM MyTable")
            try:
                print ("Or query view with CAST or TRIM to avoid trailing space padding:")
                cur.execute ("SELECT c1, c2 FROM MyView")
                [ print (row) for row in cur.fetchall () ]
            finally:
                cur.execute ("DROP VIEW MyView")
        finally:
            cur.execute ("DROP TABLE MyTable")



# package
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "unknown"

snowfinch_version = __version__
__author__ = 'gangadhar kadam'



"""
Functions for exception manipulation + custom exceptions used by SnowFinch
to identify common deviations from the expected behavior.
"""
import functools
import logging
import sys
import traceback
from typing import Optional, cast

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import EntryPoint  # pragma: no cover
else:
    from importlib_metadata import EntryPoint  # pragma: no cover

from snowfinch import __version__ as VERSION


def exceptions2exit(exception_list):
    """Decorator to convert given exceptions to exit messages
    This avoids displaying nasty stack traces to end-users
    Args:
        exception_list [Exception]: list of exceptions to convert
    """

    def exceptions2exit_decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except tuple(exception_list) as ex:
                from snowfinch.cli import get_log_level  # defer circular imports to avoid errors

                if get_log_level() <= logging.DEBUG:
                    # user surely wants to see the stacktrace
                    traceback.print_exc()
                print(f"ERROR: {ex}")
                sys.exit(1)

        return func_wrapper

    return exceptions2exit_decorator


class DirectoryAlreadyExists(RuntimeError):
    """The project directory already exists, but no ``update`` or ``force``
    option was used.
    """


class DirectoryDoesNotExist(RuntimeError):
    """No directory was found to be updated."""


class GitNotInstalled(RuntimeError):
    """PyScaffold requires git to run."""

    DEFAULT_MESSAGE = "Make sure git is installed and working."

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class GitNotConfigured(RuntimeError):
    """PyScaffold tries to read user.name and user.email from git config."""

    DEFAULT_MESSAGE = (
        "Make sure git is configured. Run:\n"
        '  git config --global user.email "you@example.com"\n'
        '  git config --global user.name "Your Name"\n'
        "to set your account's default identity."
    )

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class GitDirtyWorkspace(RuntimeError):
    """Workspace of git is empty."""

    DEFAULT_MESSAGE = (
        "Your working tree is dirty. Commit your changes first" " or use '--force'."
    )

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super().__init__(message, *args, **kwargs)


class InvalidIdentifier(RuntimeError):
    """Python requires a specific format for its identifiers.

    https://docs.python.org/3.6/reference/lexical_analysis.html#identifiers
    """


class ShellCommandException(RuntimeError):
    """Outputs proper logging when a ShellCommand fails"""


class ImpossibleToFindConfigDir(RuntimeError):
    """An expected error occurred when trying to find the config dir.

    This might be related to not being able to read the $HOME env var in Unix
    systems, or %USERPROFILE% in Windows, or even the username.
    """

    def __init__(self, message=None, *args, **kwargs):
        message = message or self.__class__.__doc__
        super().__init__(message, *args, **kwargs)



# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to create and call a SQL stored procedure
# with a variety of parameters and dynamic result sets.

import teradatasql

def DisplayResults (cur):
    while True:
        print (" === metadata ===")
        print ("  cur.rowcount:", cur.rowcount)
        print ("  cur.description:", cur.description)

        print (" === result ===")
        [ print (" ", row) for row in cur.fetchall () ]

        if not cur.nextset ():
            break

def cur_execute (cur, sSQL, params=None):
    print ()
    print ("cur.execute", sSQL, "bound values", params)
    cur.execute (sSQL, params)
    DisplayResults (cur)

def cur_callproc (cur, sProcName, params=None):
    print ()
    print ("cur.callproc", sProcName, "bound values", params)
    cur.callproc (sProcName, params) # OUT parameters are not supported by .callproc
    DisplayResults (cur)

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:

        # Demonstrate a stored procedure having IN and INOUT parameters.
        # Returns one result set having one row and one column containing the output parameter value.

        cur.execute ("replace procedure examplestoredproc (in p1 integer, inout p2 integer) begin set p2 = p1 + p2 ; end ;")
        try:
            cur_execute (cur, "{call examplestoredproc (3, 5)}") # literal parameter values
            cur_execute (cur, "{call examplestoredproc (?, ?)}", [10, 7]) # bound parameter values
            cur_callproc (cur, "examplestoredproc", [20, 4]) # bound parameter values

            # Demonstrate a stored procedure having one OUT parameter.
            # Returns one result set having one row and one column containing the output parameter value.
            # Only demonstrate .execute because OUT parameters are not supported by .callproc
            # OUT parameters must be unbound.

            cur.execute ("replace procedure examplestoredproc (out p1 varchar(100)) begin set p1 = 'foobar' ; end ;")
            cur_execute (cur, "{call examplestoredproc (?)}")

            # Demonstrate a stored procedure having no parameters that returns one dynamic result set.
            # Returns two result sets.
            # The first result set is empty having no rows or columns, because there are no output parameter values.
            # The second result set is the dynamic result set returned by the stored procedure.

            cur.execute ("""replace procedure examplestoredproc ()
                dynamic result sets 1
                begin
                    declare cur1 cursor with return for select * from dbc.dbcinfo order by 1 ;
                    open cur1 ;
                end ;""")
            cur_execute (cur, "{call examplestoredproc}")
            cur_callproc (cur, "examplestoredproc")

            # Demonstrate a stored procedure having IN and INOUT parameters that returns two dynamic result sets.
            # Returns three result sets.
            # The first result set has one row and one column containing the output parameter values.
            # The second and third result sets are dynamic result sets returned by the stored procedure.

            cur.execute ("""replace procedure examplestoredproc (in p1 integer, inout p2 integer, inout p3 integer)
                dynamic result sets 2
                begin
                    declare cur1 cursor with return for select * from dbc.dbcinfo order by 1 ;
                    declare cur2 cursor with return for select infodata, infokey from dbc.dbcinfo order by 1 ;
                    open cur1 ;
                    open cur2 ;
                    set p2 = p1 + p2 ;
                    set p3 = p1 * p3 ;
                end ;""")
            cur_execute (cur, "{call examplestoredproc (2, 1, 3)}") # literal parameter values
            cur_execute (cur, "{call examplestoredproc (?, ?, ?)}", [3, 2, 4]) # bound IN and INOUT parameter values
            cur_callproc (cur, "examplestoredproc", [10, 3, 2]) # bound IN and INOUT parameter values

            # Demonstrate a stored procedure having IN, INOUT, and OUT parameters that returns two dynamic result sets.
            # Returns three result sets.
            # The first result set has one row and two columns containing the output values from the INOUT and OUT parameters.
            # The second and third result sets are dynamic result sets returned by the stored procedure.
            # Only demonstrate .execute because OUT parameters are not supported by .callproc
            # OUT parameters must be unbound.

            cur.execute ("""replace procedure examplestoredproc (in p1 integer, inout p2 integer, out p3 varchar(100))
                dynamic result sets 2
                begin
                    declare cur1 cursor with return for select * from dbc.dbcinfo order by 1 desc ;
                    declare cur2 cursor with return for select infodata, infokey from dbc.dbcinfo order by 1 ;
                    open cur1 ;
                    open cur2 ;
                    set p2 = p1 + p2 ;
                    set p3 = 'hello' ;
                end ;""")
            cur_execute (cur, "{call examplestoredproc (10, 5, ?)}") # literal parameter values
            cur_execute (cur, "{call examplestoredproc (?, ?, ?)}", [20, 7]) # bound IN and INOUT parameter values

        finally:
            cur.execute ("drop procedure examplestoredproc")



import os
import boto3


def s3_uploader(localdir, s3dir, s3bucket, s3profile):
    boto3.setup_default_session(profile_name=s3profile)
    client = boto3.client('s3')

    # enumerate local files recursively
    for root, dirs, files in os.walk(localdir):
        for filename in files:
            # construct the full local path
            local_path = os.path.join(root, filename)
            print("local path", local_path)

            # construct the full  path
            relative_path = os.path.relpath(local_path, localdir)
            print("relative path", relative_path)
            s3_path = os.path.join(s3dir, relative_path)

            # relative_path = os.path.relpath(os.path.join(root, filename))

            print('Searching "%s" in "%s"' % (s3_path, s3bucket))
            try:
                client.head_object(Bucket=s3bucket, Key=s3_path)
                print("Path found on S3! Skipping %s..." % s3_path)

                # try:
                # client.delete_object(Bucket=bucket, Key=s3_path)
                # except:
                # print "Unable to delete %s..." % s3_path
            except Exception as ex:
                print("Uploading %s...", s3_path, ex)
                client.upload_file(local_path, s3bucket, s3_path)


# if __name__ == '__main__':
# # initiate the configuration
# configs = confuse.Configuration('SnowFinch', __name__)
#
# # Add conf items from specified file
# configs.set_file('/Users/ag29266/Downloads/pims/snowfinch-vbc-dev.yaml')
# profile = f"{configs['profile']}"
#
# s3_bucket = configs[profile]['s3']['s3_bucket'].get()
# s3_key = configs[profile]['s3']['s3_key'].get()
# s3_profile = configs[profile]['s3']['s3_profile'].get()
# local_dir = configs[profile]['s3']['localdir'].get()
#
# # s3_bucket_name = 'antm-481935479534-ssm-dev-filetransfer'
# # s3_key = 'SnowFinch/'
# # local_directory = '/Users/ag29266/Downloads/pims'
#
# # configs = configparser.RawConfigParser()
# # path = pathlib.PosixPath('~/.aws/credentials')
# # configs.read(path.expanduser())
# # profile = 'aws-dev-vbc'
# # upload_to_aws(outfile, s3_bucket_name, s3_key, configs, profile)
#
# # get an access token, local (from) directory, and S3 (to) directory
# # from the command-line
# upload_dir_to_s3(local_dir, s3_key, s3_bucket, s3_profile)



import os
import time
import hashlib
import base64

from Crypto.Cipher import AES

from snowfinch.compat import *


__all__ = ['create_key_file', 'generate_key', 'Crypto']


def create_key_file(path):
    """
    Creates a new encryption key in the path provided and sets the file
    permissions.  Setting the file permissions currently does not work
    on Windows platforms because of the differences in how file
    permissions are read and modified.
    """
    iv = "{}{}".format(os.urandom(32), time.time())
    new_key = generate_key(ensure_bytes(iv))
    with open(path, "wb") as f:
        f.write(base64.b64encode(new_key))
    os.chmod(path, 0o400)


def generate_key(s):
    for i in range(65536):
        s = hashlib.sha256(s).digest()
    return s


class Crypto(object):
    def __init__(self, key, bs=16):
        self.key = hashlib.sha256(key).digest()
        self.bs = bs

    def decrypt(self, enc):
        enc = base64.b64decode(enc)
        iv = enc[:self.bs]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self.unpad(cipher.decrypt(enc[self.bs:])).decode("utf-8")

    def encrypt(self, raw):
        raw = self.pad(raw).encode("utf-8")
        iv = os.urandom(self.bs)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(raw))

    def pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    def unpad(self, s):
        return s[:-ord(s[len(s)-1:])]

    @classmethod
    def from_key_file(cls, path):
        with open(path, "rb") as f:
            return cls(base64.b64decode(f.read()))


import os
import sys
import threading


def stringify(s):
    """In Py3k, unicode are strings, so we mustn't encode it.
    However it is necessary in Python 2.x, since Unicode strings are
    unicode, not str."""
    if type(s) != str and type(s) != bytes:
        s = s.encode('utf-8')
    return s


class LenWrapperStream(object):
    """
    A simple class to wrap a stream and provide length capability
    for streams like cStringIO

     We do it because requests will try to fallback to chunked transfer if
     it can't extract the len attribute of the object it gets, and S3 doesn't
     support chunked transfer.
     In some cases, like cStringIO, it may cause some issues, so we wrap the
     stream with a class of our own, that will proxy the stream and provide a
     proper len attribute
    """

    def __init__(self, stream):
        """
        Creates a new wrapper from the given stream

        Params:
            - stream    The baseline stream

        """
        self.stream = stream

    def read(self, n=-1):
        """
        Proxy for reading the stream
        """
        return self.stream.read(n)

    def __iter__(self):
        """
        Proxy for iterating the stream
        """
        return self.stream

    def seek(self, pos, mode=0):
        """
        Proxy for the `seek` method of the underlying stream
        """
        return self.stream.seek(pos, mode)

    def tell(self):
        """
        Proxy for the `tell` method of the underlying stream
        """
        return self.stream.tell()

    def __len__(self):
        """
        Calculate the stream length in a fail-safe way
        """
        o = self.stream

        # If we have a '__len__' method
        if hasattr(o, '__len__'):
            return len(o)

        # If we have a len property
        if hasattr(o, 'len'):
            return o.len

        # If we have a fileno property
        # (EAFP here, because some file-like objects like
        # tarfile "ExFileObject" will pass a hasattr test
        # but still not work)
        try:
            return os.fstat(o.fileno()).st_size
        except (IOError, AttributeError):
            # fallback to the manual way,
            # this is useful when using something like BytesIO
            pass

        # calculate based on bytes to end of content
        # get our start position
        start_pos = o.tell()
        # move to the end
        o.seek(0, os.SEEK_END)
        # Our len is end - start position
        size = o.tell() - start_pos
        # Seek the stream back to the start position
        o.seek(start_pos)
        # Return the size
        return size

    def __eq__(self, other):
        """
        Make sure equal method works as expected (comparing the underlying
        stream and not the wrapper)
        """
        if self.stream == other:
            return True

        if isinstance(other, LenWrapperStream) and other.stream == self.stream:
            return True

    @property
    def closed(self):
        """
        Proxy for the underlying stream closed property
        """
        return self.stream.closed

    def __repr__(self):
        """
        Proxy for the repr of the stream
        """
        return repr(self.stream)


class ProgressPercentage(object):
    """
    ProgressPercentage class is used by the S3Transfer upload_file callback
    Please see the following url for more information:
    http://boto3.readthedocs.org/en/latest/reference/customizations/s3.html#ref-s3transfer-usage
    """

    def __init__(self, filename):
        """
        Initiate the ProgressPercentage class, using the base information which
        makes up a pipeline
        Args:
            filename (str): A name of the file which we will monitor the
            progress of
        """
        self._filename = filename
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount):
        # To simplify we'll assume this is hooked up
        # to a single filename.
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write(
                "\rTransferring [{0}] {1:.2f}%".format("#" * int(percentage / 10), percentage)
            )
            sys.stdout.flush()



# Copyright 2020 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to FastExport rows from a table.

import teradatasql

with teradatasql.connect (host="whomooz", user="guest", password="please") as con:
    with con.cursor () as cur:
        with con.cursor () as cur2:
            sTableName = "FastExportTable"
            try:
                sRequest = "DROP TABLE " + sTableName
                print (sRequest)
                cur.execute (sRequest)
            except Exception as ex:
                print ("Ignoring", str (ex).split ("\n") [0])

            sRequest = "CREATE TABLE " + sTableName + " (c1 INTEGER NOT NULL, c2 VARCHAR(10))"
            print (sRequest)
            cur.execute (sRequest)

            try:
                sInsert = "INSERT INTO " + sTableName + " VALUES (?, ?)"
                print (sInsert)
                cur.execute (sInsert, [
                    [1, None],
                    [2, "abc"],
                    [3, "def"],
                    [4, "mno"],
                    [5, None],
                    [6, "pqr"],
                    [7, "uvw"],
                    [8, "xyz"],
                    [9, None],
                ])

                sSelect = "{fn teradata_try_fastexport}SELECT * FROM " + sTableName
                print (sSelect)
                cur.execute (sSelect)
                [ print (row) for row in sorted (cur.fetchall ()) ]

                sRequest = "{fn teradata_nativesql}{fn teradata_get_warnings}" + sSelect
                print (sRequest)
                cur2.execute (sRequest)
                [ print (row) for row in cur2.fetchall () ]

                sRequest = "{fn teradata_nativesql}{fn teradata_get_errors}" + sSelect
                print (sRequest)
                cur2.execute (sRequest)
                [ print (row) for row in cur2.fetchall () ]

                sRequest = "{fn teradata_nativesql}{fn teradata_logon_sequence_number}" + sSelect
                print (sRequest)
                cur2.execute (sRequest)
                [ print (row) for row in cur2.fetchall () ]

            finally:
                sRequest = "DROP TABLE " + sTableName
                print (sRequest)
                cur.execute (sRequest)



from __future__ import print_function

import os.path
import sys

import sqlparse
from ddlparse import DdlParse

from snowfinch.constants import *
from snowfinch.logger import log
from snowfinch.utils import *

logger = log.get_logger()


def parse_lines(in_filename, out_filename, start_kw, end_kw, skip_kw):
    try:
        inp = open(in_filename, 'r', encoding='utf-8', errors='ignore')
        out = open(out_filename, 'w+', encoding='utf-8', errors='ignore')
    except FileNotFoundError as err:
        print(f"File {in_filename} not found", err)
        raise
    except OSError as err:
        print(f"OS error occurred trying to open {in_filename}", err)
        raise
    except Exception as err:
        print(f"Unexpected error opening {in_filename} is",  repr(err))
        raise
    else:
        with inp, out:
            copy = False
            for line in inp:
                if not any(w in line.strip().lower() for w in skip_kw):
                    # first IF block to handle if the start and end on same line
                    if line.lstrip().lower().startswith(start_kw) and line.rstrip().endswith(end_kw):
                        copy = True
                        if copy:  # keep the starts with keyword
                            out.write(line)
                        copy = False
                        continue
                    elif line.lstrip().lower().startswith(start_kw):
                        copy = True
                        if copy:  # keep the starts with keyword
                            out.write(line)
                        continue
                    elif line.rstrip().endswith(end_kw):
                        if copy:  # keep the ends with keyword
                            out.write(line)
                        copy = False
                        continue
                    elif copy:
                        # write
                        out.write(line)


def get_volatile_table_list(filename):
    vt_list = []
    vt_kw = ["create"]
    # bad_keys = ['create', 'volatile', 'table', 'multiset', 'set', 'no fallback', ',']
    # re_bad_keys = re.compile(r"\b(" + "|".join(bad_keys) + ")\\W", re.I)
    # print(re_bad_keys)
    with open(filename, 'r', encoding='utf-8', errors='ignore') as inpt:
        for line in inpt:
            if any(line.lower().lstrip().startswith(kw) for kw in vt_kw):
                split_words = line.strip().upper().split()
                line = split_words[split_words.index('TABLE') + 1]
                vt_list.append(line.strip())
    return list(set(vt_list))


def get_source_table_list(filename):
    table_list = []
    stmt_kw = ["merge", "insert", "update", "delete"]

    with open(filename, 'r', encoding='utf-8', errors='ignore') as inpt:
        for line in inpt:
            if any(line.lower().lstrip().startswith(kw) for kw in stmt_kw):
                if line.lower().lstrip().startswith("update"):
                    split_words = line.strip().upper().split()
                    update_table = split_words[split_words.index('UPDATE') + 1]\
                        .replace("(", "")\
                        .replace(";", "")
                    table_list.append(update_table.strip().upper())
                elif line.lower().lstrip().startswith("delete from"):
                    split_words = line.strip().upper().split()
                    del_from = split_words[split_words.index('FROM') + 1]\
                        .replace("(", "")\
                        .replace(";", "")
                    table_list.append(del_from.strip().upper())
                elif line.lower().lstrip().startswith("insert"):
                    split_words = line.strip().upper().split()
                    print(split_words)
                    insert_into = split_words[split_words.index('INTO') + 1]\
                        .replace("(", "")\
                        .replace(";", "")
                    table_list.append(insert_into.strip().upper())
                elif line.lower().lstrip().startswith("merge into"):
                    split_words = line.strip().upper().split()
                    merge_into = split_words[split_words.index('INTO') + 1]\
                        .replace("(", "")\
                        .replace(";", "")
                    table_list.append(merge_into.strip().upper())
    remove_table_list = ['$', 'LL', 'LLK']
    table_list = [sa for sa in table_list if not any(sb in sa for sb in remove_table_list)]
    return list(set(table_list))


def get_table_name(line=None, stmt=None, schma=None):
    table_name = None
    full_table_name = None
    if line.lower().lstrip().startswith(stmt):
        split_words = line.strip().upper().split()

        # get table name
        if stmt.lower() in ('drop', 'create'):
            table_name = split_words[split_words.index('TABLE') + 1]\
                .replace("(", "")\
                .replace(";", "")

        if stmt.lower() == 'update':
            split_words = line.strip().upper().split()
            table_name = split_words[split_words.index('UPDATE') + 1] \
                .replace("(", "") \
                .replace(";", "")

        elif stmt.lower() == 'delete':
            split_words = line.strip().upper().split()
            table_name = split_words[split_words.index('FROM') + 1] \
                .replace("(", "") \
                .replace(";", "")

        elif stmt.lower() in ('merge', 'insert'):
            split_words = line.strip().upper().split()
            print(split_words)
            table_name = split_words[split_words.index('INTO') + 1] \
                .replace("(", "") \
                .replace(";", "")

        # assign schema
        if table_name.startswith(tuple(STG_SCHEMA_PRFX_LIST)):
            schema = f"{schma}_STG"
        elif table_name.endswith(tuple(QA_SCHEMA_PRFX_LIST)):
            schema = f"{schma}_QADATA"
        else:
            schema = schma

        # get the full table b=name
        full_table_name = f"{schema}.{table_name}"
    return [table_name, full_table_name]


def parse_sql(_input, _output, sfdb, sfschema):
    try:
        inp = open(_input, 'r', encoding='utf-8', errors='ignore')
        out = open(_output, 'w+', encoding='utf-8', errors='ignore')
    except FileNotFoundError as err:
        print(f"File {_input} not found", err)
        raise
    except OSError as err:
        print(f"OS error occurred trying to open {input}", err)
        raise
    except Exception as err:
        print(f"Unexpected error opening {input} is", repr(err))
        raise
    else:
        with inp, out:
            statements = sqlparse.split(inp.read())
            for statement in statements:
                statement = sqlparse.format(statement, strip_comments=True)
                if statement.lower().startswith('database'):
                    statement = f"USE DATABASE {sfdb};\n"

                if statement.lstrip().lower().startswith("delete from") \
                        and statement.lower().rstrip().endswith("all;"):
                    statement = statement.upper().replace("ALL", "")

                if 'sel ' in statement.lstrip().lower():
                    if statement.lstrip().lower().startswith('sel cast(count(*)'):
                        statement = "--" + statement.strip()
                    else:
                        statement = statement.lstrip().upper().replace("SEL", "SELECT")

                if 'del ' in statement.strip().lower():
                    statement = statement.lstrip().upper().replace("DEL", "DELETE")

                if statement.lstrip().lower().startswith('drop'):
                    drop_table_name = get_table_name(statement, 'drop', sfschema)
                    statement = f"\nDROP TABLE {drop_table_name[1]};\n"

                if statement.lstrip().lower().startswith('insert '):
                    insert_table_name = get_table_name(statement, 'insert', sfschema)
                    statement = statement.replace(insert_table_name[0], insert_table_name[1])

                if statement.lstrip().lower().startswith('update'):
                    update_table_name = get_table_name(statement, 'update', sfschema)
                    statement = statement.replace(update_table_name[0], update_table_name[1])

                if statement.lower().startswith('create'):
                    create_table_name = get_table_name(statement, 'create', sfschema)

                    # get DDL
                    line = sqlparse.format(statement, strip_comments=True)
                    parser = DdlParse()
                    parser.ddl = line.upper()

                    #  remove the teradata junk on create
                    parsed_ddl = parser.ddl.split("(", 1)[1:]
                    parsed_ddl1 = {
                        x.replace('CHARACTER SET LATIN NOT CASESPECIFIC', '')
                        .replace("FORMAT 'YYYY-MM-DD'", '')
                        for x in parsed_ddl
                    }
                    parsed_ddl_str = " ".join(str(x) for x in parsed_ddl1)
                    statement = f"\nCREATE TEMPORARY TABLE {create_table_name[1]} \n(" + parsed_ddl_str

                    if 'unique primary' in statement.strip().lower():
                        statement = re.sub(r'unique primary index.*;', ';', statement.strip().lower(), flags=re.DOTALL)\
                            .upper()\
                            .replace('\n;', ';')
                    if 'primary' in statement.strip().lower():
                        statement = re.sub(r'primary.*;', ';', statement.strip().lower(), flags=re.DOTALL)\
                            .upper()\
                            .replace('\n;', ';')
                        # statement = statement.upper().replace("PRIMARY INDEX", ";\n/*PRIMARY INDEX")
                    if 'index' in statement.strip().lower():
                        statement = re.sub(r'index.*;', ';', statement.strip().lower(), flags=re.DOTALL)\
                            .upper()\
                            .replace('\n;', ';')
                        # statement = statement.upper().replace("INDEX", ";\n/*INDEX")

                    # removing compress junk from the DDL
                    stmt = []
                    for i, line in enumerate(statement.splitlines()):
                        if 'COMPRESS' in line:
                            # print("comp")
                            if i == len(statement.splitlines()) - 1:
                                line = line[:line.index('COMPRESS')] + ');'
                            else:
                                line = line[:line.index('COMPRESS')] + ','
                        elif '\'' in line:
                            line = "".rstrip()
                        else:
                            line = line
                        stmt.append(line.strip())
                    statement = " \n".join(str(x) for x in stmt)
                    print(statement)

                if statement.endswith(';'):
                    statement = "\n" + statement + "\n"
                else:
                    statement = statement
                # finally write it to output file
                out.write(statement)
        # finally remove the temp file if any
        silent_remove(_input)


def build_sql(inpt_file, conf=None):
    sfdb = conf['snowflake']['database'].get()
    sfschema = conf['snowflake']['schema'].get()

    if os.path.isdir(inpt_file):
        print("input path is a directory")
        for root, dirs, files in os.walk(inpt_file):
            for file in files:
                in_filename = root + '/' + file
                print('Working on', in_filename)
                if is_source(in_filename):
                    # out_filename = in_filename + '.sql'
                    out_raw_file = get_outfile_full_name(in_filename, 'raw')
                    out_sql_file = get_outfile_full_name(in_filename, 'sql')

                    print('Generating the output sql...')
                    # process_bteq_scripts(in_filename, out_filename, skip_lines_list)

                    # raw
                    parse_lines(in_filename, out_raw_file, tuple(STMT_START_KW), tuple(STMT_END_KW), STMT_SKIP_KW)

                    # final sql
                    parse_sql(out_raw_file, out_sql_file, sfdb, sfschema)

                    print('conversion completed successfully\n')
    elif os.path.isfile(inpt_file):
        print("input path is a single file")
        out_raw_file = get_outfile_full_name(inpt_file, 'raw')
        out_sql_file = get_outfile_full_name(inpt_file, 'sql')

        # process_bteq_scripts(inpt_file, inpt_file + '.sql', skip_lines_list)
        # outpt_file = get_outfile_full_name(inpt_file)

        # raw
        parse_lines(inpt_file, out_raw_file, tuple(STMT_START_KW), tuple(STMT_END_KW), STMT_SKIP_KW)

        # final sql
        parse_sql(out_raw_file, out_sql_file, sfdb, sfschema)
    else:
        print('Not a file or directory', inpt_file, file=sys.stderr)
        sys.exit(-1)



import pandas as pd
import yaml
from sqlalchemy import create_engine
from snowflake.sqlalchemy import URL
import boto3
from io import StringIO
from io import BytesIO


def read_sql(sql, conn):
    try:
        # read data using a sql query
        df = pd.read_sql_query(sql, conn)

        # trim columns
        df = df.applymap(lambda x: x.strip() if isinstance(x, str) else x)

        # upper case columns
        df.columns = [x.upper() for x in df.columns]

    except Exception as err:
        print('Error occurred while executing a query {}'.format(err.args))

    return df


def load_yaml(conf_file):
    with open(conf_file, "r") as cf:
        return yaml.full_load(cf)


def pad_left(n, width, pad="0"):
    return ((pad * width) + str(n))[-width:]


def copy_to_s3(df, bucket, filepath, fmt, s3profile):
    boto3.setup_default_session(profile_name=s3profile)
    client = boto3.client('s3')

    if fmt == 'parquet':
        out_buffer = BytesIO()
        df.to_parquet(out_buffer, index=False, compression='snappy')

    elif fmt == 'csv':
        out_buffer = StringIO()
        df.to_csv(out_buffer, header=True, index=False)

    out_buffer.seek(0)

    # csv_buf = StringIO()
    # df.to_csv(csv_buf, header=True, index=False)
    # csv_buf.seek(0)

    client.put_object(Bucket=bucket, Body=out_buffer.getvalue(), Key=filepath)
    print(f'Copy {df.shape[0]} rows to S3 Bucket {bucket} at {filepath}, Done!')


if __name__ == '__main__':
    config = load_yaml('/Users/ag29266/PycharmProjects/cloudflow/config/pt-inntwk-wgs.yaml')
    # Set options below

    conn = create_engine(URL(
        account='eda_prod.anthemdatalake.us-east-1.privatelink',
        user='AN118647AD',
        database='P01_EDL',
        schema="EDL_ALLPHI",
        warehouse="P01_EDL_USER_WH_M",
        authenticator="externalbrowser",
        validate_default_parameters=True
    ))
    # ein = read_sql(q, conn)
    # ein.to_csv(r'/Users/ag29266/Downloads/MBRSHP_WGS_WGMGRPR.csv', index=False, header=True)
    # ep_rltnshp = read_sql("select * from ep_rltnshp", conn)
    # print(ep_rltnshp.head())

    # copy_to_s3(ep_rltnshp,
    #            'antm-sf-sit-dataz-nogbd-phi-useast1',
    #            'cnfz/edlc01/prov/edw/ep_rltnshp/mbrshp_wgs_wgmgrpr.parquet',
    #            'parquet',
    #            'aws-dev-pt')

    count = 0
    table_name = "sps_cd_val_trnsltn"
    query = "select * from sps_cd_val_trnsltn"

    s3_source_bucket = "antm-sf-sit-dataz-nogbd-phi-useast1"
    s3_source_path = "cnfz/edlc01/clm/edw"

    for chunk in pd.read_sql_query(query, conn, chunksize=1500000):
        copy_to_s3(chunk,
                   s3_source_bucket,
                   f's3_source_path/{table_name}/snowfinch_part_%s.snappy.parquet' % pad_left(count, 4),
                   'parquet',
                   'aws-prod-exec')
        count += 1












from __future__ import absolute_import

import os
import ntpath
import errno
from typing import List
import re


def file_exists(path):
    try:
        with open(path):
            return True
    except (OSError, IOError):
        return False


def file_permissions(path):
    return os.stat(path).st_mode & 0o777


def home_file(filename):
    return os.path.join(os.path.expanduser("~"), filename)


def isfile(s):
    if isinstance(s, (str, bytes)) and ' ' not in s and file_exists(s):
        return True
    return False


def is_source(filename):
    suffixes = ('.sh', '.hql', '.sql', '.txt')
    for s in suffixes:
        if filename.endswith(s):
            return True
    return False


def get_outfile_name(infile, extension):
    filename = ntpath.basename(infile)
    out_file_name = filename.split(".")[0] + extension
    return out_file_name


def get_outfile_full_name(infile, extension):
    filename = ntpath.basename(infile)
    filepath = ntpath.dirname(infile)
    out_file_name = filename.split(".")[0]
    outfilename = f"{filepath}/{out_file_name}.{ extension}"
    return outfilename


def silent_remove(file):
    try:
        os.remove(file)
    except OSError as e:
        if e.errno != errno.ENOENT:  # no such file or directory
            raise  # re-raise exception if a different error occurred


def get_version_info():
    from snowfinch.__init__ import __version__
    return __version__


def import_string(import_name):
    import_name = str(import_name)
    if not import_name.startswith("snowfinch."):
        import_name = "snowfinch.{}".format(import_name)
    try:
        return __import__(import_name)
    except ImportError:
        try:
            module_name, obj_name = import_name.rsplit('.', 1)
        except AttributeError as error:
            raise ImportError(error)
        try:
            module = __import__(module_name, None, None, [obj_name])
        except ImportError:
            module = import_string(module_name)
        try:
            return getattr(module, obj_name)
        except AttributeError as error:
            raise ImportError(error)


def regex(value):
    """ Helper for regex types on the command line. """
    try:
        return re.compile(value)
    except Exception as ex:
        print(ex)
        raise ValueError


def get_file_paths_list(local_dir: str) -> List[str]:
    """
    Create list of file paths.
    :param local_dir: Local filepath of assets to SCP to host.
    :type local_dir: List[str]
    """
    local_files = os.walk(local_dir)
    for root, dirs, files in local_files:
        return [f"{local_dir}/{file}" for file in files]



import os
from pathlib import PurePath
from .exceptions import DBError, S3CredentialsError
from .aws.s3 import S3
from .logger import log

logger = log.get_logger()

COPY_FORMAT_OPTIONS = {
    "csv": {
        "compression",
        "record_delimiter",
        "field_delimiter",
        "skip_header",
        "date_format",
        "time_format",
        "timestamp_format",
        "binary_format",
        "escape",
        "escape_unenclosed_field",
        "trim_space",
        "field_optionally_enclosed_by",
        "null_if",
        "error_on_column_count_mismatch",
        "validate_utf8",
        "empty_field_as_null",
        "skip_byte_order_mark",
        "encoding",
    },
    "json": {
        "compression",
        "file_extension",
        "enable_octal",
        "allow_duplicate",
        "strip_outer_array",
        "strip_null_values",
        "ignore_utf8_errors",
        "skip_byte_order_mark",
    },
    "parquet": {"binary_as_text"},
}


UNLOAD_FORMAT_OPTIONS = {
    "csv": {
        "compression",
        "record_delimiter",
        "field_delimiter",
        "file_extension",
        "date_format",
        "time_format",
        "timestamp_format",
        "binary_format",
        "escape",
        "escape_unenclosed_field",
        "field_optionally_enclosed_by",
        "null_if",
    },
    "json": {"compression", "file_extension"},
    "parquet": {"snappy_compression"},
}


def combine_options(options=None):
    """ Returns the ``copy_options`` or ``format_options`` attribute with spaces in between and as
    a string. If options is ``None`` then return an empty string.
    """
    return " ".join(options) if options is not None else ""


class Fincher(S3):
    def __init__(self, profile=None, conn=None):
        self.conn = conn
        try:
            S3.__init__(self, profile)
        except S3CredentialsError:
            logger.warning("S3 credentials were not found. S3 functionality is disabled")
            logger.warning("Only internal stages are available")

    def upload_to_stage(self, local, stage, parallel=4, auto_compress=True, overwrite=True):
        local_uri = PurePath(local).as_posix()
        self.conn.execute(
            "PUT 'file://{0}' {1} PARALLEL={2} AUTO_COMPRESS={3} OVERWRITE={4}".format(
                local_uri, stage, parallel, auto_compress, overwrite
            )
        )

    def download_from_stage(self, stage, local=None, parallel=10):
        if local is None:
            local = os.getcwd()
        local_uri = PurePath(local).as_posix()
        self.conn.execute("GET {0} 'file://{1}' PARALLEL={2}".format(stage, local_uri, parallel))

    def load(self, table_name, stage, file_type="csv", format_options=None, copy_options=None):
        if not self.conn:
            raise DBError("No Snowflake connection object is present.")

        if file_type not in COPY_FORMAT_OPTIONS:
            raise ValueError(
                "Invalid file_type. Must be one of {0}".format(list(COPY_FORMAT_OPTIONS.keys()))
            )

        if format_options is None and file_type == "csv":
            format_options = ["FIELD_DELIMITER='|'", "SKIP_HEADER=0"]

        format_options_text = combine_options(format_options)
        copy_options_text = combine_options(copy_options)
        base_copy_string = "COPY INTO {0} FROM '{1}' " "FILE_FORMAT = (TYPE='{2}' {3}) {4}"
        try:
            sql = base_copy_string.format(
                table_name, stage, file_type, format_options_text, copy_options_text
            )
            self.conn.execute(sql, commit=True)

        except Exception as e:
            logger.error("Error running COPY on Snowflake. err: %s", e)
            raise DBError("Error running COPY on Snowflake.")

    def unload(
        self,
        stage,
        table_name,
        file_type="csv",
        format_options=None,
        header=False,
        copy_options=None,
    ):

        if not self.conn:
            raise DBError("No Snowflake connection object is present")

        if file_type not in COPY_FORMAT_OPTIONS:
            raise ValueError(
                "Invalid file_type. Must be one of {0}".format(list(UNLOAD_FORMAT_OPTIONS.keys()))
            )

        if format_options is None and file_type == "csv":
            format_options = ["FIELD_DELIMITER='|'"]

        format_options_text = combine_options(format_options)
        copy_options_text = combine_options(copy_options)
        base_unload_string = (
            "COPY INTO {0} FROM {1} " "FILE_FORMAT = (TYPE='{2}' {3}) HEADER={4} {5}"
        )

        try:
            sql = base_unload_string.format(
                stage, table_name, file_type, format_options_text, header, copy_options_text
            )
            self.conn.execute(sql, commit=True)
        except Exception as e:
            logger.error("Error running UNLOAD on Snowflake. err: %s", e)
            raise DBError("Error running UNLOAD on Snowflake.")



# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to insert a batch of rows.

import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        cur.execute ("create volatile table voltab (c1 integer, c2 varchar(100)) on commit preserve rows")

        cur.execute ("insert into voltab (?, ?)", [
            [1, "abc"],
            [2, "def"],
            [3, "ghi"]])

        cur.execute ("select * from voltab order by 1")
        [ print (row) for row in cur.fetchall () ]



from snowfinch.logger import log

logger = log.get_logger()


def get_ddl(engine, db, schema, table):
    ddl = None
    logger.info(f"Fetching DDL from {engine}")
    if engine.lower() == 'mssql':
        ddl = get_ddl_mssql(db, schema, table)
    elif engine.lower() == 'teradata':
        ddl = get_ddl_tera(db, schema, table)
    else:
        print("DB engine not  supported yet")
    return ddl


def get_sproc_def(engine, db, schema, sp_name):
    sp_source_text = None
    logger.info(f"Fetching DDL from {engine}")
    if engine.lower() == 'mssql':
        sp_source_text = get_sproc_def_mssql(db, schema, sp_name)
    elif engine.lower() == 'teradata':
        sp_source_text = get_sproc_def_tera(db, sp_name)
    else:
        print("DB engine not  supported yet")
    return sp_source_text


def get_ddl_tera(db, schema, table):
    print(db)
    ddl_stmt = f"""
with snowfinch_ddl as(
select 
C.databasename,
C.tablename,
columntype,
case columntype 
when 'CF' then 'CHAR'
when 'CV' then 'VARCHAR'
when 'DA' then 'DATE'
when 'I8' then 'BIGINT'
when 'TS' then 'TIMESTAMP'
when 'D' then 'DECIMAL'
when 'I' then 'INTEGER'
when 'I2' then 'SMALLINT' 
when 'I1' then 'BYTEINT'
when 'N' then 'NUMBER' 
when 'F' then 'FLOAT' 
end as dtype,
case 
when columntype in ('CF','CV') then '('||trim(cast(columnlength as integer))||')'
when columntype in ('D','N') then '('||trim(cast(decimaltotaldigits as integer))||','||trim(cast(decimalfractionaldigits as integer))||')'
when columntype in ('TS') then '('||trim(cast(decimalfractionaldigits as integer))||')'
else ''
end dlen,
 case nullable when 'N' then 'NOT NULL' else '' end as nulltype,
 case 
 when columntype in ('CF','CV','I8','I','I2','D','N','F','I1') and defaultvalue is not null then 'DEFAULT '||trim(defaultvalue) 
 when columntype in ('DA') and defaultvalue not like all('%date','null') then 'DEFAULT '''||trim(defaultvalue)||trim('''::DATE') 
 when columntype in ('DA') and defaultvalue like ('%date') then 'DEFAULT '||trim('CURRENT_DATE')||trim('::DATE') 
 else '' 
 end as deftype,
 max(columnid) over (partition by C.databasename,C.tablename order by columnid) as maxid,
 min(columnid) over (partition by C.databasename,C.tablename order by columnid) as minid,
 case when maxid=columnid then trim(coalesce(I.keycolumns,''))||' );' else ',' end as commadelim,
 case when minid=columnid and T.commitopt='N' and C.databasename not like all('%temp%','%data%') then 'CREATE TABLE IF NOT EXISTS {schema}.'||upper(C.tablename)||' ( '
 when minid=columnid and T.commitopt='N' and C.databasename like any('%temp%','%data%') then 'CREATE TABLE IF NOT EXISTS {schema}.'||upper(C.tablename)||' ( '
 when minid=columnid and T.commitopt in ('P','D') then 'CREATE TABLE IF NOT EXISTS {schema}.'||upper(C.tablename)||' ( 'else '' end as createtbl,
 case 
 when idcoltype in ('GD','GA') then 'AUTOINCREMENT' else '' 
 end as autoinc,
 columnid,
 dense_rank() over (order by C.tablename) TBL_COUNT,
 case when T.commitopt in ('P','D') then 'Temporary' else 'Permanent' end as tbltype,
 createtbl||columnname||' '||trim(dtype)||trim(dlen)||' '||trim(nulltype)||' '||trim(deftype)||' '||trim(autoinc)||' '||trim(commadelim) as ddl
from dbc.columnsv C
join dbc.tablesv T
on C.databasename=T.databasename
and C.tablename=T.tablename
and T.tablekind IN('T','O')
left outer join 
(select databasename,tablename,COALESCE(', CONSTRAINT PK_'||trim(tablename)||' PRIMARY KEY ('||TRIM(TRAILING ',' FROM XMLAGG(columnName || ',' ORDER BY columnPosition)(varchar(4000)))||' )','') AS KeyColumns
from dbc.indicesv
group by 1,2
where indextype='K' or uniqueflag='Y'
) I
on I.databasename=T.databasename
and I.tablename=T.tablename
where 
--upper(c.databasename) in ('PI','RDM','ETL_TEMP_PI','ETL_DATA_PI','EDW_V20','QADATA_PI')
upper(c.databasename) in ('{db}')
and upper(c.tablename)='{table}'
)

sel DDL
from snowfinch_ddl
order by tbltype,databasename,tablename,columnid;
 """
    # print(ddl_stmt)
    return ddl_stmt


def get_ddl_mssql(db, schema, table):
    print(db)

    ddl_stmt = f"""
DECLARE @table_name SYSNAME
        SELECT @table_name = 'dbo.{table}'
        DECLARE 
              @object_name SYSNAME
            , @object_id INT
        SELECT 
              @object_name = '' + s.name + '.' + o.name + ''
            , @object_id = o.[object_id]
        FROM sys.objects o WITH (NOWAIT)
        JOIN sys.schemas s WITH (NOWAIT) ON o.[schema_id] = s.[schema_id]
        WHERE s.name + '.' + o.name = @table_name
            AND o.[type] = 'U'
            AND o.is_ms_shipped = 0

        DECLARE @SQL NVARCHAR(MAX) = ''

        ;WITH index_column AS 
        (
            SELECT 
                  ic.[object_id]
                , ic.index_id
                , ic.is_descending_key
                , ic.is_included_column
                , c.name
            FROM sys.index_columns ic WITH (NOWAIT)
            JOIN sys.columns c WITH (NOWAIT) ON ic.[object_id] = c.[object_id] AND ic.column_id = c.column_id
            WHERE ic.[object_id] = @object_id
        ),
        fk_columns AS 
        (
             SELECT 
                  k.constraint_object_id
                , cname = c.name
                , rcname = rc.name
            FROM sys.foreign_key_columns k WITH (NOWAIT)
            JOIN sys.columns rc WITH (NOWAIT) ON rc.[object_id] = k.referenced_object_id AND rc.column_id = k.referenced_column_id 
            JOIN sys.columns c WITH (NOWAIT) ON c.[object_id] = k.parent_object_id AND c.column_id = k.parent_column_id
            WHERE k.parent_object_id = @object_id
        )
        SELECT @SQL = 'CREATE TABLE IF NOT EXISTS {schema}.' + substring(@object_name,5,len(@object_name)) + CHAR(13) + '(' + CHAR(13) + STUFF((
            SELECT CHAR(9) + ', ' + c.name + ' ' + 
                CASE WHEN c.is_computed = 1
                    THEN 'AS ' + cc.[definition] 
                    ELSE case when tp.name = 'datetime' then 'timestamp_ntz(9)'  else tp.name end + 
                        CASE WHEN tp.name IN ('varchar', 'char', 'varbinary', 'binary')
                               THEN '(' + CASE WHEN c.max_length = -1 THEN '16777216' ELSE CAST(c.max_length AS VARCHAR(5)) END + ')'
                             WHEN tp.name IN ('nvarchar', 'nchar')
                               THEN '(' + CASE WHEN c.max_length = -1 THEN '16777216' ELSE CAST(c.max_length / 2 AS VARCHAR(5)) END + ')'
                             WHEN tp.name IN ('datetime2', 'time2', 'datetimeoffset') 
                               THEN '(' + CAST(c.scale AS VARCHAR(5)) + ')'
                             WHEN tp.name IN ('decimal','numeric') 
                               THEN '(' + CAST(c.[precision] AS VARCHAR(5)) + ',' + CAST(c.scale AS VARCHAR(5)) + ')'
                            ELSE ''
                        END +  
                        CASE WHEN ic.is_identity = 1 THEN ' IDENTITY(' + CAST(ISNULL(ic.seed_value, '0') AS CHAR(1)) + ',' + CAST(ISNULL(ic.increment_value, '1') AS CHAR(1)) + ')' ELSE '' END + 
                        CASE WHEN c.is_nullable = 1 THEN ' NULL' ELSE ' NOT NULL' END +
                        CASE WHEN dc.[definition] IS NOT NULL THEN ' DEFAULT' + dc.[definition] ELSE '' END                
                END + CHAR(13)
            FROM sys.columns c WITH (NOWAIT)
            JOIN sys.types tp WITH (NOWAIT) ON c.user_type_id = tp.user_type_id
            LEFT JOIN sys.computed_columns cc WITH (NOWAIT) ON c.[object_id] = cc.[object_id] AND c.column_id = cc.column_id
            LEFT JOIN sys.default_constraints dc WITH (NOWAIT) ON c.default_object_id != 0 AND c.[object_id] = dc.parent_object_id AND c.column_id = dc.parent_column_id
            LEFT JOIN sys.identity_columns ic WITH (NOWAIT) ON c.is_identity = 1 AND c.[object_id] = ic.[object_id] AND c.column_id = ic.column_id
            WHERE c.[object_id] = @object_id
            ORDER BY c.column_id
            FOR XML PATH(''), TYPE).value('.', 'NVARCHAR(MAX)'), 1, 2, CHAR(9) + ' ')
            + ISNULL((SELECT CHAR(9) + ', CONSTRAINT ' + k.name + ' PRIMARY KEY  (' + 
                            (SELECT STUFF((
                                 SELECT ', ' + c.name + ' ' + CASE WHEN ic.is_descending_key = 1 THEN 'DESC' ELSE '' END
                                 FROM sys.index_columns ic WITH (NOWAIT)
                                 JOIN sys.columns c WITH (NOWAIT) ON c.[object_id] = ic.[object_id] AND c.column_id = ic.column_id
                                 WHERE ic.is_included_column = 0
                                     AND ic.[object_id] = k.parent_object_id 
                                     AND ic.index_id = k.unique_index_id     
                                 FOR XML PATH(N''), TYPE).value('.', 'NVARCHAR(MAX)'), 1, 2, ''))
                    + ')' + CHAR(13)
                    FROM sys.key_constraints k WITH (NOWAIT)
                    WHERE k.parent_object_id = @object_id 
                        AND k.[type] = 'PK'), '') + ')'  + CHAR(13)

SELECT @SQL AS DDL;
"""

    # print(ddl_stmt)
    return ddl_stmt


def get_sproc_def_mssql(_db, _schema, _sp_name):
    sproc_def = f"""
    SELECT SPECIFIC_NAME, ROUTINE_DEFINITION
    FROM {_db}.INFORMATION_SCHEMA.ROUTINES
    WHERE  
    UPPER(SPECIFIC_CATALOG) = '{_db}'
    AND UPPER(SPECIFIC_SCHEMA) = '{_schema}'
    AND UPPER(ROUTINE_TYPE) = 'PROCEDURE'
    AND UPPER(SPECIFIC_NAME) = '{_sp_name}'
    """
    return sproc_def


def get_sproc_args_mssql(_db, _schema, _sp_name):
    args_sql = f"""
    SELECT  
    P.SPECIFIC_NAME,
    UPPER(REPLACE(CONCAT(P.PARAMETER_NAME, ' ',
    CASE 
    WHEN P.DATA_TYPE = 'INT' then P.DATA_TYPE 
    WHEN P.DATA_TYPE = 'datetime' then 'timestamp_ntz(9)'  
    
    WHEN P.DATA_TYPE IN ('varchar', 'char', 'varbinary', 'binary') THEN 
        CASE 
            WHEN P.CHARACTER_MAXIMUM_LENGTH = -1 THEN  P.DATA_TYPE 
            ELSE '(' + CAST( P.CHARACTER_MAXIMUM_LENGTH AS VARCHAR(5)) + ')'
        END
    WHEN P.DATA_TYPE  IN ('nvarchar', 'nchar') THEN 
        CASE
            WHEN P.CHARACTER_MAXIMUM_LENGTH = -1 THEN P.DATA_TYPE 
            ELSE '(' + CAST( P.CHARACTER_MAXIMUM_LENGTH AS VARCHAR(5)) + ')'
        END

    ELSE P.DATA_TYPE
    END),'@','')) AS  PARAM
    FROM {_db}.INFORMATION_SCHEMA.PARAMETERS p
    WHERE 
    UPPER(P.SPECIFIC_SCHEMA) = '{_schema}' 
    AND UPPER(P.SPECIFIC_NAME) = '{_sp_name}';
    """
    return args_sql


def get_sproc_def_tera(_engine, _sp_name):
    sproc_def = f"""
    """
    return sproc_def



# Copyright 2019 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to FastLoad batches of rows.

import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        sTableName = "FastLoadBatch"
        try:
            sRequest = "DROP TABLE " + sTableName
            print (sRequest)
            cur.execute (sRequest)
        except Exception as ex:
            print ("Ignoring", str (ex).split ("\n") [0])

        sRequest = "CREATE TABLE " + sTableName + " (c1 INTEGER NOT NULL, c2 VARCHAR(10))"
        print (sRequest)
        cur.execute (sRequest)

        try:
            sRequest = "{fn teradata_nativesql}{fn teradata_autocommit_off}"
            print (sRequest)
            cur.execute (sRequest)

            aaoValues = [
                [1, # c1 INTEGER NOT NULL
                None, # c2 VARCHAR
                ],
                [2, # c1 INTEGER NOT NULL
                "abc", # c2 VARCHAR
                ],
                [3, # c1 INTEGER NOT NULL
                "def", # c2 VARCHAR
                ],
            ]

            sInsert = "{fn teradata_try_fastload}INSERT INTO " + sTableName + " (?, ?)"
            print (sInsert)
            cur.execute (sInsert, aaoValues)

            sRequest = "{fn teradata_nativesql}{fn teradata_get_warnings}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            sRequest = "{fn teradata_nativesql}{fn teradata_get_errors}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            sRequest = "{fn teradata_nativesql}{fn teradata_logon_sequence_number}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            aaoValues = [
                [4, # c1 INTEGER NOT NULL
                "mno", # c2 VARCHAR
                ],
                [5, # c1 INTEGER NOT NULL
                None, # c2 VARCHAR
                ],
                [6, # c1 INTEGER NOT NULL
                "pqr", # c2 VARCHAR
                ],
            ]

            print (sInsert)
            cur.execute (sInsert, aaoValues)

            sRequest = "{fn teradata_nativesql}{fn teradata_get_warnings}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            sRequest = "{fn teradata_nativesql}{fn teradata_get_errors}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            aaoValues = [
                [7, # c1 INTEGER NOT NULL
                "uvw", # c2 VARCHAR
                ],
                [8, # c1 INTEGER NOT NULL
                "xyz", # c2 VARCHAR
                ],
                [9, # c1 INTEGER NOT NULL
                None, # c2 VARCHAR
                ],
            ]

            print (sInsert)
            cur.execute (sInsert, aaoValues)

            sRequest = "{fn teradata_nativesql}{fn teradata_get_warnings}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            sRequest = "{fn teradata_nativesql}{fn teradata_get_errors}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            print ("con.commit()")
            con.commit ()

            sRequest = "{fn teradata_nativesql}{fn teradata_get_warnings}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            sRequest = "{fn teradata_nativesql}{fn teradata_get_errors}" + sInsert
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

            sRequest = "{fn teradata_nativesql}{fn teradata_autocommit_on}"
            print (sRequest)
            cur.execute (sRequest)

            sRequest = "SELECT * FROM " + sTableName + " ORDER BY 1"
            print (sRequest)
            cur.execute (sRequest)
            [ print (row) for row in cur.fetchall () ]

        finally:
            sRequest = "DROP TABLE " + sTableName
            print (sRequest)
            cur.execute (sRequest)



from boto3 import Session
from boto3.s3.transfer import TransferConfig
from botocore.client import Config
from snowfinch.exceptions import *
from snowfinch.aws.s3utils import ProgressPercentage
from snowfinch.constants import *
from snowfinch.logger import log

logger = log.get_logger()


class S3(object):

    def __init__(self, profile=None, kms_key=None, **kwargs):

        self.profile = profile
        self.kms_key = kms_key
        self.session = None
        self.s3 = None

        self._set_session()
        self._set_client()

    def _set_session(self):
        try:
            self.session = Session(profile_name=self.profile)
            logger.info("Initialized AWS session.")
        except Exception as e:
            logger.error("Error initializing AWS Session, err: %s", e)
            raise S3Error("Error initializing AWS Session.")
        credentials = self.session.get_credentials()
        if credentials is None:
            raise S3CredentialsError("Credentials could not be set.")

    def _set_client(self):
        try:
            self.s3 = self.session.client("s3", config=Config(signature_version="s3v4"))
            logger.info("Successfully initialized S3 client.")
        except Exception as e:
            logger.error("Error initializing S3 Client, err: %s", e)
            raise S3InitializationError("Error initializing S3 Client.")

    def _credentials_string(self):
        creds = self.session.get_credentials()
        if creds.token is not None:
            temp = "aws_access_key_id={};aws_secret_access_key={};token={}"
            return temp.format(creds.access_key, creds.secret_key, creds.token)
        else:
            temp = "aws_access_key_id={};aws_secret_access_key={}"
            return temp.format(creds.access_key, creds.secret_key)

    @staticmethod
    def _generate_s3_path(bucket, key):
        return "s3://{0}/{1}".format(bucket, key)

    @staticmethod
    def _generate_unload_path(bucket, folder):
        if folder:
            s3_path = "s3://{0}/{1}".format(bucket, folder)
        else:
            s3_path = "s3://{0}".format(bucket)
        return s3_path

    def upload_to_s3(self, local, bucket, key):
        extra_args = {}
        try:
            # force ServerSideEncryption
            if self.kms_key is None:
                extra_args["ServerSideEncryption"] = "AES256"
                logger.info("Using AES256 for encryption")
            else:
                extra_args["ServerSideEncryption"] = "aws:kms"
                extra_args["SSEKMSKeyId"] = self.kms_key
                logger.info("Using KMS Keys for encryption")

            logger.info("Uploading file to S3 bucket: %s", self._generate_s3_path(bucket, key))
            self.s3.upload_file(
                local, bucket, key, ExtraArgs=extra_args, Callback=ProgressPercentage(local)
            )
        except Exception as e:
            logger.error("Error uploading to S3. err: %s", e)
            raise S3UploadError("Error uploading to S3.")

    def upload_list_to_s3(self, local_list, bucket, folder=None):
        output = []
        for file in local_list:
            if folder is None:
                s3_key = os.path.basename(file)
            else:
                s3_key = "/".join([folder, os.path.basename(file)])
            self.upload_to_s3(file, bucket, s3_key)
            output.append("/".join([bucket, s3_key]))
        return output

    def download_from_s3(self, bucket, key, local):
        try:
            logger.info(
                "Downloading file from S3 bucket: %s", self._generate_s3_path(bucket, key),
            )
            config = TransferConfig(max_concurrency=5)
            self.s3.download_file(bucket, key, local, Config=config)
        except Exception as e:
            logger.error("Error downloading from S3. err: %s", e)
            raise S3DownloadError("Error downloading from S3.")

    def download_list_from_s3(self, s3_list, local_path=None):
        if local_path is None:
            local_path = os.getcwd()

        output = []
        for f in s3_list:
            s3_bucket, key = self.parse_s3_url(f)
            local = os.path.join(local_path, os.path.basename(key))
            self.download_from_s3(s3_bucket, key, local)
            output.append(local)
        return output

    def delete_from_s3(self, bucket, key):
        try:
            logger.info("Deleting file from S3 bucket: %s", self._generate_s3_path(bucket, key))
            self.s3.delete_object(Bucket=bucket, Key=key)
        except Exception as e:
            logger.error("Error deleting from S3. err: %s", e)
            raise S3DeletionError("Error deleting from S3.")

    def delete_list_from_s3(self, s3_list):

        for file in s3_list:
            s3_bucket, s3_key = self.parse_s3_url(file)
            self.delete_from_s3(s3_bucket, s3_key)

    @staticmethod
    def parse_s3_url(s3_url):
        temp_s3 = s3_url.replace("s3://", "")
        s3_bucket = temp_s3.split("/")[0]
        s3_key = "/".join(temp_s3.split("/")[1:])
        return s3_bucket, s3_key



import pandas as pd
import yaml
import boto3
from io import StringIO
from io import BytesIO
from sqlalchemy import create_engine
import logging

logger = logging.getLogger("tera2s3")
logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')


def td_connect(config):
    usr = config['teradata'].get('user')
    passwd = config['teradata'].get('password')
    host = config['teradata'].get('host')
    url: str = f'teradatasql://{host}/?user={usr}&password={passwd}&logmech' \
               f'=LDAP&tmode=TERA&encryptdata=true'
    print(url)
    try:
        logger.info("Connecting to Teradata database...")
        conn = create_engine(url.strip()).connect()
        logger.info("Successfully connected to Teradata.")
        return conn
    except Exception as err:
        print("Failed to connect to Teradata DB: {}.".format(err))
        raise


def pad_left(n, width, pad="0"):
    return ((pad * width) + str(n))[-width:]


def load_yaml(conf_file):
    with open(conf_file, "r") as cf:
        return yaml.full_load(cf)


def copy_to_s3(df, bucket, filepath, fmt, s3profile):
    boto3.setup_default_session(profile_name=s3profile)
    client = boto3.client('s3')
    if fmt == 'parquet':
        out_buffer = BytesIO()
        df.to_parquet(out_buffer, index=False, compression='snappy', engine='pyarrow')
    elif fmt == 'csv':
        out_buffer = StringIO()
        df.to_csv(out_buffer, header=True, index=False)
    out_buffer.seek(0)
    # csv_buf = StringIO()
    # df.to_csv(csv_buf, header=True, index=False)
    # csv_buf.seek(0)
    client.put_object(Bucket=bucket, Body=out_buffer.getvalue(), Key=filepath)
    print(f'Copy {df.shape[0]} rows to S3 Bucket {bucket} at {filepath}, Done!')


def get_counts(chunk):
    voters_street = chunk[
        "Residential Address Street Name "]
    return voters_street.value_counts()


if __name__ == '__main__':
    app_config_file = '/Users/AG29266/Library/Application Support/JetBrains/PyCharm2021.2/scratches/scratch.yml'
    config = load_yaml(app_config_file)
    tdc = td_connect(config)
    url = "jdbc:teradata://DWPROD2COP1.CORP.ANTHEM.COM/database=etl_views_prov_adl,tmode=TERA,charset=UTF8, logmech=LDAP"
    driver = "com.teradata.jdbc.TeraDriver"

    properties = {
        "user": config['teradata'].get('user'),
        "password": config['teradata'].get('password'),
        "driver": driver
    }

    sql_query = "select * from edw_allphi.sps_cd_val_trnsltn"
    table_name = "sps_cd_val_trnsltn"
    count = 0
    # for chunk in pd.read_sql_query(clm_hist_sql, tdc, chunksize=1000000):
    #     copy_to_s3(chunk,
    #                'antm-sf-sit-dataz-nogbd-phi-useast1',
    #                'cnfz/edlc01/rdm/edw/fncl_prod_cf/snowfinch_part_%s.parquet' % pad_left(count,4),
    #                'parquet',
    #                'aws-dev-pt')
    #     count += 1
    s3_source_bucket_clm = "antm-pt-sit-dataz-nogbd-nophi-useast1"
    s3_source_path_clm = "cnfz/edlc01/prov/edw/"

    for chunk in pd.read_sql_query(sql_query, tdc, chunksize=2000000):
        copy_to_s3(chunk,
                   s3_source_bucket_clm,
                   f'{s3_source_path_clm}/{table_name}/snowfinch_part_%s.snappy.parquet' % pad_left(count, 4),
                   'parquet',
                   'aws-dev-pt')
        count += 1





import os
import pathlib

# there should be a folder with the host file and the credentials path
try:
    config_folder = pathlib.Path(os.environ["SNOWFINCH_CONFIG"])
except KeyError:
    config_folder = pathlib.Path.home() / ".snowfinch"

# create folder
try:
    config_folder.mkdir(parents=True)
except OSError:
    pass

HOST_PATH = config_folder / ".hosts.json"
CREDENTIAL_PATH = config_folder / ".configs.json"

# the cryptography key can be in the same folder or at a separate location
try:
    KEY_PATH = os.environ["SNOWFINCH_KEY"]
except KeyError:
    KEY_PATH = config_folder / ".configkey"

STMT_START_KW = ['database ', 'drop ', 'create ', 'insert ', 'delete ', 'update ', 'merge ', '/*']
STMT_END_KW = ";"
STMT_SKIP_KW = ['error handling','#''collect stats proc',
                'call', '.if', '.run','lo comments end',
                'lo comments start','.export', '.quit',
                '.label', '.exit', '.set', '.run', '.logon',
                'bteq', 'set query band', 'select session',
                'call', 'sel sum(currentspool)', '.goto', 'fi',
                'rc', 'if', 'else', 'echo', 'exec', 'rm ',
                'cat', 'then', 'elif', 'exit', 'eof',
                '. $code', '#', 'awk ', 'sed ', '(echo'
                '/gpfs03', '${code}', 'export',
                ]

STG_SCHEMA_PRFX_LIST = ['WORK_', 'WRK_', 'TEMP_', 'TMP_', 'RZ_', 'VTT_', 'VT_', 'GTT', 'BLNCG', "_VT", "_VTT"]
QA_SCHEMA_PRFX_LIST = ['_LOAD_LOG', '_CNTRL']

# database flavors
DB_FLAVORS = (
    'postgresql',
    'mysql',
    'mssql',
    'teradata',
    'snowflake',
    'redshift',)

CACHE_SIZE = 64
CHUNK_SIZE = 30000



# Copyright 2019 by Teradata Corporation. All rights reserved.

# This sample program demonstrates the behavior of the fake_result_sets escape function.

import teradatasql

with teradatasql.connect (host="whomooz", user="guest", password="please") as con:
    with con.cursor () as cur:
        cur.execute ("create volatile table voltab1 (c1 integer, c2 varchar(100)) on commit preserve rows")
        cur.execute ("insert into voltab1 values (1, 'abc')")
        cur.execute ("insert into voltab1 values (2, 'def')")

        cur.execute ("create volatile table voltab2 (c1 integer, c2 varchar(100)) on commit preserve rows")
        cur.execute ("insert into voltab2 values (3, 'ghi')")
        cur.execute ("insert into voltab2 values (4, 'jkl')")
        cur.execute ("insert into voltab2 values (5, 'mno')")

        cur.execute ("create volatile table voltab3 (c1 integer, c2 varchar(100)) on commit preserve rows")
        cur.execute ("insert into voltab3 values (6, 'pqr')")
        cur.execute ("insert into voltab3 values (7, 'stu')")
        cur.execute ("insert into voltab3 values (8, 'vwx')")
        cur.execute ("insert into voltab3 values (9, 'yz')")

        cur.execute ("{fn teradata_fake_result_sets}select * from voltab1 order by 1")
        print ("=== Two result sets produced by a single-statement SQL request that returns one result set ===")
        print (" --- Fake result set ---")
        row = cur.fetchone ()
        [ print (" Column {} {:15} = {}".format (i + 1, cur.description [i][0], row [i])) for i in range (0, len (row)) ]
        cur.nextset ()
        print (" --- Real result set ---")
        [ print (row) for row in cur.fetchall () ]
        print ()

        cur.execute ("{fn teradata_fake_result_sets}select * from voltab2 order by 1 ; select * from voltab3 order by 1")
        print ("=== Four result sets produced by a multi-statement SQL request that returns two result sets ===")
        print (" --- Fake result set ---")
        row = cur.fetchone ()
        [ print (" Column {} {:15} = {}".format (i + 1, cur.description [i][0], row [i])) for i in range (0, len (row)) ]
        cur.nextset ()
        print (" --- Real result set ---")
        [ print (row) for row in cur.fetchall () ]
        cur.nextset ()
        print (" --- Fake result set ---")
        row = cur.fetchone ()
        [ print (" Column {} {:15} = {}".format (i + 1, cur.description [i][0], row [i])) for i in range (0, len (row)) ]
        cur.nextset ()
        print (" --- Real result set ---")
        [ print (row) for row in cur.fetchall () ]



from __future__ import print_function

from snowfinch.sshclient import RemoteClient
from snowfinch.parsers.bteqparser import *
from snowfinch.parsers.ddlparser import *
from snowfinch.parsers.sproc_parser import *
from snowfinch.connections import snowfinch_connections
from snowfinch.aws.s3upload import s3_uploader
from snowfinch.logger import log

logger = log.get_logger()


def get_s3_upload(_config):
    script_dir = _config['bteq']['script_dir'].get()
    script_name = _config['bteq']['source_files'].get()
    s3_bucket = _config['s3']['s3_bucket'].get()
    s3_key = _config['s3']['s3_key'].get()
    s3_profile = _config['s3']['s3_profile'].get()

    if script_name is None:
        logger.info(f"Uploading all SnowFinch SQL from the input dir")
        local_file = script_dir + "/*.sql"
    else:
        logger.info(f"Uploading a single Snowfinch SQL file")
        infile = script_dir + "/" + script_name

        local_file = get_outfile_full_name(infile, 'sql')

    logger.info(f"input scripts dir: {script_dir}")
    logger.info(f"input scripts name: {script_name}\n")
    logger.info(f"input scripts name and path: {local_file}\n")

    try:
        logger.info(f"Uploading the SnowFinch SQL to S3\n")
        s3_uploader(local_file, s3_key, s3_bucket, s3_profile)
        logger.info(f"Successfully uploaded the SnowFinch SQL files to S3\n")
    except Exception as ex:
        logger.error(ex)


def get_scp_to_local(host, user, passwd, ssh_key, remote_dir, download_dir, remote_files):
    # get ssh client
    client = RemoteClient(host, user, passwd, ssh_key, remote_dir)
    # download the files from remote to local
    with client.scp as scp:
        # Downloading a file from remote machine
        for file in remote_files:
            remote_file = remote_dir + file
            local_file = download_dir + file
            scp.get(remote_file, local_file)


def get_scp_to_remote(host, user, passwd, ssh_key, local_dir, upload_dir, local_files):
    client = RemoteClient(host, user, passwd, ssh_key, upload_dir)
    with client.scp as scp:
        # Downloading a file from remote machine
        # Uploading file from local to remote machine
        for local_file in local_files:
            local_file = local_dir + local_file
            scp.put(local_file, upload_dir)


def get_bteq_build(_config):
    script_dir = _config['bteq']['script_dir'].get()
    script_name = _config['bteq']['source_files'].get()
    logger.info(f"input scripts dir: {script_dir}")
    logger.info(f"input scripts name: {script_name}\n")
    if script_name is None:
        infile = script_dir
    else:
        infile = script_dir + "/" + script_name
    logger.info(f"infile: {infile}")
    build_sql(infile, _config)


def get_ddl_build(dialect, _config, drop_table=False):
    # get the target(snowflake)  connection objects
    tgt_conn = snowfinch_connections("snowflake", _config)

    # get the source(teradata,mssql)  connection objects
    src_conn = get_source_db_connections(dialect, _config)

    src_table_list = _config[dialect]['tablelist'].get()

    # get_ddl_conversion(tgt_conn, src_conn, configs, td_table_list)
    build_table(dialect, src_conn, tgt_conn, src_table_list, _config, drop_table=drop_table)


def get_filenames(directory, exclude=None, include=None, concurrency=1,  split=False):
    """
    Return a list of filenames to upload, filtered by :attr:`include` and
    :attr:`exclude`, if set.

    If `split` is ``True``, then this method returns a list of lists, where
    filenames are evenly divided into :attr:`concurrency` groups.

    After running this method, :attr:`total` will be set to the number of
    filenames found.

    """
    filenames = []
    total = 0
    for path, dirs, files in os.walk(directory):
        for filename in files:
            filename = os.path.join(path, filename)
            # Iterate over all the include regexes, determining if we
            # should include this filename
            if include:
                skip = True
                for reg in include:
                    if reg.search(filename):
                        skip = False
                        break
                if skip:
                    continue
            # Iterate over the exclude regexes, seeing if we should skip
            if exclude:
                skip = False
                for reg in exclude:
                    if reg.search(filename):
                        skip = True
                        break
                if skip:
                    continue
            filenames.append(filename)
            total += 1

    if split:
        groups = [list() for _ in range(concurrency)]
        for i in range(len(filenames)):
            groups[i % concurrency].append(filenames[i])
        filenames = groups

    return filenames


def get_sproc_build(dialect, _config):
    if dialect == 'mssql':
        build_sproc_mssql(_config)
    else:
        logger.info("Stored procedure currently supported for msql only")


def get_full_build(dialect, _config, drop_table=False):
    script_source = _config['bteq']['source'].get()
    script_dir = _config['bteq']['script_dir'].get()
    script_names = _config['bteq']['source_files'].get()
    logger.info(f"input scripts dir: {script_dir}")
    infile_list = None
    if script_names is None:
        logger.info(f"input scripts list is  empty")
        logger.info(f"input scripts source is {script_source}")

        if script_source.lower() == "remote":
            logger.info(f"remote input scripts dir: {script_dir}")

            # get ssh config
            local_dir = _config['bteq']['download_dir'].get()
            host_name = _config['bteq']['remote_host'].get()
            user_name = _config['bteq']['remote_user'].get()
            pass_word = _config['bteq']['remote_password'].get()
            ssh_key = _config['bteq']['ssh_key_path'].get()
            remote_files = "{local_dir}/*"
            # download the remote files to local
            get_scp_to_local(host_name, user_name, pass_word, ssh_key, script_dir, local_dir, remote_files)
            # shutil.rmtree(dirpath)
        else:
            logger.info(f"input scripts source is {script_source}")
            infile_list = get_filenames(script_dir)
    else:
        logger.info(f"input scripts files: {script_names}")
        infile_list = [os.path.join(script_dir, name) for name in script_names]

    logger.info(infile_list)

    logger.info("creating source and target connections ...\n")
    tgt_conn = snowfinch_connections("snowflake", _config)

    # get the source  connection objects
    src_conn = get_source_db_connections(dialect, _config)

    logger.info("\nConverting Teradata BTEQ to SnowSQL...")

    for file in infile_list:
        logger.info(f"\ninput script to be converted: {file}")

        # get the VT list
        volatile_table_list = get_volatile_table_list(file)
        print("\nVolatile Tables List....")
        print(*volatile_table_list, sep='\n')

        # Get the source tables
        source_table_list = get_source_table_list(file)
        print("\nSource Tables List...")
        print(*source_table_list, sep='\n')

        # get only not  vt tables
        if volatile_table_list:
            logger.info("Volatile tables found, removing them from exist check..")
            src_table_list = list(set(source_table_list) - set(volatile_table_list))
        else:
            logger.info("No Volatile tables seen in the input script")
            src_table_list = source_table_list

        print("\nFinal Table List...")
        print(*src_table_list, sep='\n')

        if src_table_list:
            build_table(dialect, src_conn, tgt_conn, src_table_list, _config, drop_table=drop_table)
        else:
            logger.info("All tables are volatile table, Skipping the table exist")

        # build snowflake sql from Teradata
        build_sql(file, _config)

        # finally, upload the sql to s3
        get_s3_upload(_config)



import sqlalchemy as sa

from snowfinch.constants import *
from snowfinch.connections import *
from snowfinch.parsers.commons import *
from snowfinch.parsers.ddl_generators import *
from snowfinch.exceptions import *

logger = log.get_logger()


def process_snowflake_ddl(df):
    # make column names to uppercase
    df.columns = [x.upper() for x in df.columns]

    df['DDL'] = df['DDL'].str \
        .replace('   ,', ',', regex=True) \
        .replace('   ', ' ', regex=True)
    return df


def drop_table_if_exists(conn, table, schema):
    if check_table_exists(conn, table, schema):
        logger.info("Table exists on snowflake, going ahead with  drop table")

        try:
            conn.execute(f"drop table if exists {schema}.{table}")
            logger.info(f"Successfully dropped the table {table}")
        except ProgrammingError as exc:
            logger.info(f"*"*40)
            logger.warn(exc)
            logger.warn(f"Insufficient privileges to operate on {table}")
            logger.info(f"*"*40)
        except DatabaseError as err:
            logger.error(err)
            logger.info(f"*"*40)
            logger.error(f"Failed dropping the table {table}")
            logger.info(f"*"*40)

    else:
        logger.info("Table does not exists on snowflake, skipping the table drop")


def check_table_exists(conn, table, schema):
    # db = conf['snowflake']['database'].get()
    # schema = conf['snowflake']['schema'].get()
    ins = sa.inspect(conn)
    ret = ins.dialect.has_table(conn.connect(), f"{schema}.{table}")
    print('Table "{}" exists: {}'.format(table, ret))
    return ret


def build_table(dialect, src_conn, tgt_conn, table_list, conf, drop_table=False):
    # get snowflake  parameters
    sf_user = conf['snowflake']['user'].get()
    sf_role = conf['snowflake']['role'].get()
    sf_wh = conf['snowflake']['warehouse'].get()
    sf_db = conf['snowflake']['database'].get()
    sf_table_owner_role = conf['snowflake']['owner_role'].get()
    sf_schema = conf['snowflake']['schema'].get()
    src_db = conf[dialect]['database'].get()
    add_audit_cols = conf['snowflake']['add_audit_cols'].get()
    obj_typ = 'table'

    try:
        with src_conn as s_conn, tgt_conn.connect() as t_conn:
            for table in table_list:
                if "." in table:
                    src_db = table.split('.')[0]
                    src_table = table.split('.')[1]
                else:
                    src_db = src_db
                    src_table = table

                if src_table.startswith(tuple(STG_SCHEMA_PRFX_LIST)):
                    _sf_schema = f"{sf_schema}_STG"
                elif src_table.endswith(tuple(QA_SCHEMA_PRFX_LIST)):
                    _sf_schema = f"{sf_schema}_QADATA"
                else:
                    _sf_schema = sf_schema

                logger.info(f"\nCreating {src_table} on Snowflake @{_sf_schema}...")
                logger.info("snowflake table: " + src_table)
                logger.info("snowflake database: " + sf_db)
                logger.info("snowflake schema: " + _sf_schema)

                # table exist
                table_exists = check_table_exists(t_conn, src_table, _sf_schema)
                logger.info("checking if table exist...")
                logger.info(f"*" * 40)
                if table_exists:
                    logger.warn(f"table {src_table} already exists in snowflake!!")
                    if drop_table:
                        logger.warn(f"table exists but drop_table is true!!")
                        logger.warn(f"dropping the table...")
                        logger.info("")
                        logger.info("checking the table grants...")
                        owner = get_object_owner(t_conn, src_table, obj_typ, _sf_schema)
                        logger.info(f"owner: {owner}")
                        logger.info("")
                        if not (owner == sf_role or owner == sf_table_owner_role):
                            logger.warn(f"user {sf_user} does not have the privileges to drop the table")
                            logger.warn(f"please check with {owner} to drop this table!!")
                        else:
                            logger.info("dropping the table...")
                            drop_table_if_exists(t_conn, src_table, _sf_schema)
                            logger.info("")
                    else:
                        logger.warn(f"table exists but drop_table is false!!")
                        logger.warn(f"drop_table=False, hence skipping the ddl creation!!")
                        logger.info("")
                else:
                    logger.info(f"table {src_table} doesn't exists in snowflake!!")
                    logger.info(f"generating the ddl for  {src_table}...")

                    # Generate DDL
                    logger.info("generating the table ddl...")
                    logger.info(f"*" * 40)
                    ddl_sql = get_ddl(dialect, src_db, _sf_schema, src_table)

                    # get the SnowFlaked DDL
                    raw_df = read_sql(ddl_sql, s_conn)
                    logger.info("ddl row count: ", len(raw_df.index))
                    logger.info("checking is the table exist on source...")
                    logger.info("")
                    if len(raw_df.index) == 0:
                        logger.error(f"{src_table} does not exist on source DB, Please check!!\n")
                        logger.info("")
                    else:
                        # transform the DDL
                        logger.info("transforming the ddl...")
                        final_df = process_snowflake_ddl(raw_df)
                        logger.info(final_df)
                        logger.info("")

                        # convert DDL to a list
                        logger.info("converting the ddl to a list...")
                        lst1 = final_df['DDL'].astype(str).tolist()
                        logger.info(lst1)
                        logger.info("")

                        audit_col_list = [
                            'EDL_RUN_ID VARCHAR(255) NOT NULL COLLATE \'utf8\' DEFAULT \'NA\',',
                            'EDL_SCRTY_LVL_CD VARCHAR(10) NOT NULL COLLATE \'utf8\' DEFAULT \'NA\',',
                            'EDL_LOB_CD VARCHAR(10) NOT NULL COLLATE \'utf8\' DEFAULT \'NA\',',
                            'EDL_EXTRNL_LOAD_CD VARCHAR(10) NOT NULL COLLATE \'utf8\' DEFAULT \'NA\',',
                            'EDL_SOR_CD VARCHAR(10) NOT NULL COLLATE \'utf8\' DEFAULT \'NA\',',
                            'EDL_LOAD_DTM TIMESTAMP_NTZ(9) NOT NULL DEFAULT CAST(\'8888-12-31 00:00:00\' AS '
                            'TIMESTAMP_NTZ(9)) ,'
                        ]
                        if add_audit_cols == "Y":
                            logger.info("add_audit_cols=Y, Adding EDL audit columns...")
                            _index = lst1[0].index('(') + 1
                            l1 = [f'CREATE TABLE IF NOT EXISTS {_sf_schema}.{src_table} (']
                            logger.info(l1)
                            # remove the l1 from first element
                            lst1[0] = lst1[0][_index:]
                            # copy the list
                            lst2 = lst1[:]
                            logger.info(lst2)
                            # add the audit columns to the list2
                            lst2[0:0] = audit_col_list[:]
                            logger.info(lst2)
                            # lst3 = lst2[:]
                            # lst3[0] = l1 + lst
                            lst = l1 + lst2
                        else:
                            logger.info("Audit Columns Flag is N, Skipping the audit column additions")
                            lst = lst1

                        # clean the create table statement
                        create_table_stmt = '\n'.join(map(str, lst))
                        logger.info("create_table_stmt....")
                        logger.info(create_table_stmt)

                        # drop table - optional
                        # if drop_table:
                        #     logger.info("check the table grants")
                        #     # grant_ownership(src_table, _sf_schema, sf_table_owner_role, t_conn)
                        #     owner = get_object_owner(t_conn, src_table, obj_typ, _sf_schema)
                        #     logger.info(owner)
                        #     if not (owner == sf_role or owner == sf_table_owner_role):
                        #         logger.info(f"user {sf_user} does not have the privileges to drop the table")
                        #     else:
                        #         drop_table_if_exists(t_conn, src_table, _sf_schema)

                        # Execute the create table statement
                        logger.info("executing the ddl on snowflake...")
                        logger.info(f"*" * 40)
                        t_conn.execute(f'use warehouse {sf_wh}')
                        t_conn.execute(f'use database {sf_db}')
                        t_conn.execute(f'use schema {_sf_schema}')
                        logger.info(f"Creating {src_table}")
                        t_conn.execute(f"""
                              {create_table_stmt}
                          """)
                        # Change the ownership of the table
                        grant_ownership(t_conn, src_table, obj_typ, _sf_schema, sf_table_owner_role)
                        # Check the table metadata to confirm  table creation and ownership
                        om = get_object_metadata(t_conn, src_table, obj_typ, _sf_schema)
                        logger.info(om)
                        logger.info(f"Successfully created {src_table} on SnowFlake\n")
    except Exception as exc:
        logger.exception(exc)



# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to use the Elicit File feature of the Teradata SQL Driver for Python
# to upload the C source file to the Teradata Database and create a User-Defined Function (UDF).
# This sample program requires the udfinc.c source file to be located in the current directory.

import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        print ('Create function')
        cur.execute ("create function myudfinc(integer) returns integer language c no sql parameter style sql external name 'CS!udfinc!udfinc.c!F!udfinc'")

        print ('Execute function')
        cur.execute ("select myudfinc(1)")

        print ('Function returned', cur.fetchone () [0])

        print ('Drop function')
        cur.execute ("drop function myudfinc")



# Copyright 2019 by Teradata Corporation. All rights reserved.

# This sample program measures the time to insert one million rows as 100 batches of 10000 rows per batch.

import sys
import teradatasql
import time

sTableName = "batchinsperf"
nRowsPerBatch = 10000
nBatchCount = 100

with teradatasql.connect (host="whomooz", user="guest", password="please") as con:

    with con.cursor () as cur:

        try:
            cur.execute ("drop table " + sTableName)
        except Exception as ex:
            print ("Ignoring", str (ex).split ("\n") [0])

        cur.execute ("create table " + sTableName + " (c1 integer, c2 varchar(100), c3 integer, c4 varchar(100))")
        try:
            dStartTime1 = time.time ()

            for iBatch in range (0, nBatchCount):

                aaoBatch = []
                for nRow in range (1, nRowsPerBatch + 1):
                    nValue = iBatch * nRowsPerBatch + nRow
                    aaoBatch += [[nValue, "a" + str (nValue), -nValue, "b" + str (nValue)]]

                dStartTime2 = time.time ()

                cur.execute ("insert into " + sTableName + " values (?, ?, ?, ?)", aaoBatch)

                print ("Batch insert #{} of {} rows took {:.0f} ms".format (iBatch + 1, nRowsPerBatch, (time.time () - dStartTime2) * 1000))

                # end for iBatch

            dTotalTime = time.time () - dStartTime1
            print ("Inserting {} batches containing {} total rows took {:.2f} sec for {:.2f} rows/sec throughput".format (
                nBatchCount,
                nBatchCount * nRowsPerBatch,
                dTotalTime,
                nBatchCount * nRowsPerBatch / dTotalTime
            ))

            cur.execute ("select count(*) from " + sTableName + " order by 1")
            print ("Table {} contains {} rows".format (sTableName, cur.fetchone () [0]))

        finally:
            cur.execute ("drop table " + sTableName)



import inspect
import logging
import os
import platform
from colorama import Fore
from datetime import datetime, timezone
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from typing import Any, NoReturn, Union
from typing import Dict, List, Optional, Tuple

import pyfiglet
from colorama import Style

from snowfinch import snowfinch_version

LOG_ATTRIBUTES_TO_NAME_AND_FORMAT_AND_COLOR_DICT = {
    "asctime": ("s", Fore.YELLOW),
    "created": ("f", Fore.YELLOW),
    "msecs": ("d", Fore.YELLOW),
    "relativeCreated": ("d", Fore.YELLOW),
    "name": ("s", Fore.MAGENTA),
    "levelname": ("s", Fore.RED),
    "levelno": ("d", Fore.RED),
    "pathname": ("s", Fore.CYAN),
    "filename": ("s", Fore.CYAN),
    "lineno": ("d", Fore.CYAN),
    "module": ("s", Fore.BLUE),
    "funcName": ("s", Fore.BLUE),
    "process": ("d", Fore.LIGHTGREEN_EX),
    "processName": ("s", Fore.LIGHTGREEN_EX),
    "thread": ("d", Fore.LIGHTBLUE_EX),
    "threadName": ("s", Fore.LIGHTBLUE_EX),
    "exc_info": (None, Fore.BLACK),
    "exc_text": (None, Fore.BLACK),
    "stack_info": (None, Fore.BLACK),
    "args": (None, Fore.WHITE),
}
# "msg" and "message" are excluded on purpose

DEFAULT_IGNORE_ATTRIBUTE_LIST = [
    "args",
    "created",
    "exc_info",
    "exc_text",
    "pathname",
    "levelno",
    "msecs",
    "relativeCreated",
    "stack_info",
    "module",
    "funcName",
    "thread",
    "threadName",
    "process",
    "processName",
]


def get_apache_log_format(
        attr_config_dict: Dict[str, Tuple[Optional[str], Optional[str], int]],
        ignore_attr_list: List[str],
        add_colors: bool = False,
) -> str:
    format_list = list()
    for attr, (str_format, color) in attr_config_dict.items():

        if attr in ["message", "msg"] or attr in ignore_attr_list:
            continue

        # attr_name = attr if attr_name is None else attr_name
        str_format = "s" if str_format is None else str_format
        if add_colors is True:
            format_list.append(
                f"{Fore.LIGHTBLACK_EX}{Style.RESET_ALL} "
                + f"{color}%({attr}){str_format}{Style.RESET_ALL}:".strip()
            )
        else:
            format_list.append(f"%({attr}){str_format}")

    format_list.append("%(message)s")

    return " ".join(format_list)


class Apache(logging.Formatter):
    converter = datetime.fromtimestamp

    def __init__(self, fmt=None, datefmt=None, style="%", **kwargs):

        self.use_utc = kwargs.pop("use_utc", False)
        self.add_colors = kwargs.pop("add_colors", False)
        self.ignore_log_attribute_list = kwargs.pop("ignore_log_attribute_list", None)
        self.timezone = (
            timezone.utc if self.use_utc is True else datetime.now().astimezone().tzinfo
        )

        if self.ignore_log_attribute_list is None:
            self.ignore_log_attribute_list = DEFAULT_IGNORE_ATTRIBUTE_LIST

        if fmt is not None:
            log_format = fmt
        else:
            log_format: str = get_apache_log_format(
                attr_config_dict=LOG_ATTRIBUTES_TO_NAME_AND_FORMAT_AND_COLOR_DICT,
                ignore_attr_list=self.ignore_log_attribute_list,
                add_colors=self.add_colors,
            )
        super().__init__(fmt=log_format, datefmt=datefmt, style=style)

    def formatTime(self, record, datefmt=None):
        """Override: logger.Formatter.formatTime"""
        ct = self.converter(record.created)
        if datefmt is not None:
            s = ct.strftime(datefmt)
        else:
            s = datetime.fromtimestamp(record.created, tz=self.timezone)
        return s


class InvalidValue(Exception):
    def __init__(self, value: Any, allowed_value_list: List[Any]) -> NoReturn:
        self.value = value
        self.allowed_value_list = allowed_value_list

    def __str__(self) -> str:
        return f"'{self.value}' is not one of the allowed values: {self.allowed_value_list}."


class SnowFinchLogger(object):
    ALLOWED_FORMATTER_STR_LIST = ["json", "apache"]

    def __init__(
            self,
            project_name: str = "SnowFinch",
            log_level: str = "info",
            assign_logger_name: bool = False,
            formatter: Union[str, logging.Formatter] = "apache",
            log_to_stdout: bool = True,
            log_to_file: bool = False,
            log_dir: str = "logs",
            log_file_suffix: str = "S",
            rotate_file_by_size: bool = False,
            rotating_file_max_size_bytes: int = 1048576,
            rotate_file_by_time: bool = False,
            rotation_period: str = "H",
            rotation_interval: int = 1,
            rotation_time: datetime.time = None,
            rotating_file_backup_count: int = 1024,
            use_utc: bool = False,
            colors_to_stdout: bool = True,
            ignore_log_attribute_list: List[str] = None,
    ) -> NoReturn:

        self.project_name = project_name
        self.log_level = log_level.upper()
        self.assign_logger_name = assign_logger_name
        self.input_formatter = formatter
        self.log_to_stdout = log_to_stdout
        self.log_to_file = log_to_file
        self.log_dir = os.path.abspath(log_dir)
        self.log_file_suffix = log_file_suffix
        self.rotate_file_by_size = rotate_file_by_size
        self.rotating_file_max_size_bytes = rotating_file_max_size_bytes
        self.rotate_file_by_time = rotate_file_by_time
        self.rotation_period = rotation_period.upper()
        self.rotation_interval = rotation_interval
        self.rotation_time = rotation_time
        self.rotating_file_backup_count = rotating_file_backup_count
        self.use_utc = use_utc
        self.colors_to_stdout = colors_to_stdout
        self.ignore_log_attribute_list = ignore_log_attribute_list
        self.__set_logger()
        self.__set_log_handlers()
        self.__print_project_name()

    def __set_logger(self) -> NoReturn:
        if self.assign_logger_name is True:
            self.__logger = logging.getLogger(name=self.project_name)
        else:
            self.__logger = logging.getLogger()

        self.__logger.setLevel(self.log_level)

    def get_logger(self) -> logging.Logger:
        return self.__logger

    def __set_log_handlers(self) -> NoReturn:
        if self.rotate_file_by_size is True:
            self.__set_log_filepath(set_suffix=True)
            fh = RotatingFileHandler(
                filename=self.__log_filepath,
                encoding="utf-8",
                maxBytes=self.rotating_file_max_size_bytes,
                backupCount=self.rotating_file_backup_count,
            )
            self.__add_handler(handler=fh, add_colors=False)

        elif self.rotate_file_by_time is True:
            self.__validate_rotation_period()
            self.__set_log_filepath(set_suffix=False)
            fh = TimedRotatingFileHandler(
                filename=self.__log_filepath,
                encoding="utf-8",
                when=self.rotation_period,
                interval=self.rotation_interval,
                backupCount=self.rotating_file_backup_count,
                utc=self.use_utc,
                atTime=self.rotation_time,
            )
            self.__add_handler(handler=fh, add_colors=False)

        elif self.log_to_file is True:
            self.__set_log_filepath(set_suffix=True)
            fh = logging.FileHandler(filename=self.__log_filepath, encoding="utf-8")
            self.__add_handler(handler=fh, add_colors=False)

        else:
            pass

        if self.log_to_stdout is True:
            sh = logging.StreamHandler()
            self.__add_handler(handler=sh, add_colors=self.colors_to_stdout)

    def __set_log_filepath(self, set_suffix: bool = False) -> NoReturn:
        self.__set_log_filename(set_suffix=set_suffix)
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)
        self.__log_filepath = os.path.join(self.log_dir, self.__log_filename)

    def __validate_rotation_period(self) -> NoReturn:
        allowed_rotation_period_list = ["S", "M", "H", "D", "MIDNIGHT"]
        allowed_rotation_period_list += list(map(lambda n: f"W{n}", range(0, 7)))
        if self.rotation_period not in allowed_rotation_period_list:
            raise InvalidValue(
                self.rotation_period,
                allowed_rotation_period_list,
            )

    def __set_log_filename(self, set_suffix: bool = False) -> NoReturn:
        if set_suffix is True:
            self.__log_filename = "{0}_{1}.{2}".format(
                self.project_name, self.__get_log_filename_suffix(), "log"
            )
        else:
            self.__log_filename = f"{self.project_name}.log"

    def __get_log_filename_suffix(self) -> str:
        suffix_to_date_time_format_dict = {
            "S": "%Y-%m-%d_%H-%M-%S",
            "M": "%Y-%m-%d_%H-%M-00",
            "H": "%Y-%m-%d_%H-00-00",
            "D": "%Y-%m-%d",
        }

        if self.log_file_suffix not in suffix_to_date_time_format_dict.keys():
            raise InvalidValue(
                self.log_file_suffix, list(suffix_to_date_time_format_dict.keys())
            )
        datetime_now = self.__get_datetime_now()

        return datetime_now.strftime(
            suffix_to_date_time_format_dict[self.log_file_suffix]
        )

    def __get_datetime_now(self) -> datetime:
        if self.use_utc is True:
            return datetime.utcnow()
        else:
            return datetime.now()

    def __add_handler(self, handler: logging.Handler, add_colors: bool) -> NoReturn:
        handler.setLevel(self.log_level)
        handler.setFormatter(self.__get_formatter(add_colors=add_colors))
        self.__logger.addHandler(handler)

    def __get_formatter(self, add_colors: bool) -> logging.Formatter:
        if isinstance(self.input_formatter, logging.Formatter) is True:
            return self.input_formatter

        elif isinstance(self.input_formatter, str) is True:
            # if self.input_formatter.lower() == "json":
            #     return Json(
            #         use_utc=self.use_utc,
            #         ignore_log_attribute_list=self.ignore_log_attribute_list,
            #     )
            if self.input_formatter.lower() == "apache":
                return Apache(
                    use_utc=self.use_utc,
                    add_colors=add_colors,
                    ignore_log_attribute_list=self.ignore_log_attribute_list,
                )
            else:
                raise InvalidValue(
                    value=self.input_formatter,
                    allowed_value_list=self.ALLOWED_FORMATTER_STR_LIST,
                )
        else:
            raise TypeError("'formatter' must be of type 'logger.Formatter' or 'str'")

    def __print_project_name(self) -> NoReturn:
        ascii_text = pyfiglet.figlet_format(self.project_name, font="slant")
        print(f"Welcome to\n{ascii_text}version {snowfinch_version}")
        print("Using Python version %s (%s, %s)" % (
            platform.python_version(),
            platform.python_build()[0],
            platform.python_build()[1]))
        print("Supported dialects: 'Teradata', 'Microsoft SQL Server', 'SnowFlake'.")

    def log_function_call(self, func):
        def wrapper(*args, **kwargs):
            func_args = inspect.signature(func).bind(*args, **kwargs).arguments
            self.get_logger().info(
                "{0}.{1}({2})".format(
                    func.__module__,
                    func.__qualname__,
                    ", ".join("{} = {!r}".format(*item) for item in func_args.items()),
                )
            )
            return func(*args, **kwargs)

        return wrapper


log = SnowFinchLogger(
    project_name="SnowFinch",
    log_to_stdout=False,
    formatter="apache",
    log_to_file=True,
    log_dir="/Users/ag29266/Downloads/snowfinch/logs"
)



"""
Functions for exception manipulation + custom exceptions used by SnowFinch
to identify common deviations from the expected behavior.
"""
import functools
import logging
import sys
import traceback


class SnowFinchBaseError(Exception):
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors


class ConfigurationError(SnowFinchBaseError):
    """
    For use with configuration file handling.
    """
    pass


class FileNotFound(SnowFinchBaseError):
    """
    Raised when file does not exist.
    """
    pass


class ConfigNotFound(ConfigurationError, FileNotFound):
    """
    Raised when the specified configuration file does not exist.
    """
    pass


class ConnectionLock(ConfigurationError):
    """
    Raised when connection is locked by invalid attempts and the
    'protect' feature is being used.
    """

    def __init__(self, dsn):
        super(ConnectionLock, self).__init__(("Connection {0} is currently locked. please update "
                                              "credentials and run:\n\tsnowfinch config --unlock {0}").format(dsn))


class ConfigReadOnly(ConfigurationError):
    """
    Raised when a write is attempted on a configuration file was opened
    in read mode.
    """
    pass


class MissingPackage(SnowFinchBaseError):
    pass


class ProgrammingError(SnowFinchBaseError):
    pass


class DatabaseError(SnowFinchBaseError):
    pass


class OperationalError(SnowFinchBaseError):
    pass


class ShellCommandException(RuntimeError):
    """Outputs proper logging when a ShellCommand fails"""
    pass


class CompressionError(SnowFinchBaseError):
    """
    Raised when there is an error compressing a file.
    """
    pass


class ClassNotFoundException(Exception):
    pass


class ConfigError(Exception):
    pass


class ParameterError(Exception):
    pass


class FieldTypeError(Exception):
    pass


class FieldNotExistsError(Exception):
    pass


class FieldAlreadyExistsError(Exception):
    pass


class FieldTypeValueConversionError(Exception):
    pass


class MigrationNotExistsError(Exception):
    pass


class MigrationNotCompatibleError(Exception):
    pass


class ResourceNotExistsError(Exception):
    pass


class TableNotExistsError(Exception):
    pass


class DBError(Exception):
    """
    Base class for all DBSource errors.
    """


class CredentialsError(DBError):
    """
    Raised when the users credentials are not provided.
    """


class S3Error(Exception):
    """
    Base class for all S3 errors.
    """


class S3CredentialsError(S3Error):
    """
    Raised when there is an error with AWS credentials.
    """


class S3InitializationError(S3Error):
    """
    Raised when there is an error initializing S3 client.
    """


class S3UploadError(S3Error):
    """
    Raised when there is an upload error to S3.
    """


class S3DownloadError(S3Error):
    """
    Raised when there is an download error to S3.
    """


class S3DeletionError(S3Error):
    """
    Raised when there is an deletion error on S3.
    """


class InvalidIdentifier(RuntimeError):
    """Python requires a specific format for its identifiers.
    https://docs.python.org/3.6/reference/lexical_analysis.html#identifiers
    """


def exceptions2exit(exception_list):
    """Decorator to convert given exceptions to exit messages
    This avoids displaying nasty stack traces to end-users
    Args:
        exception_list [Exception]: list of exceptions to convert
    """

    def exceptions2exit_decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except tuple(exception_list) as ex:
                from snowfinch.cli import get_log_level  # defer circular imports to avoid errors

                if get_log_level() <= logging.DEBUG:
                    # user surely wants to see the stacktrace
                    traceback.print_exc()
                print(f"ERROR: {ex}")
                sys.exit(1)

        return func_wrapper

    return exceptions2exit_decorator



"""
Shell commands like git, django-admin etc.
"""

import functools
import os
import shlex
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Union

from snowfinch.exceptions import ShellCommandException
from snowfinch.logger import log

logger = log.get_logger()

PathLike = Union[str, os.PathLike]

# The following default flags were borrowed from Github's docs:
# https://docs.github.com/en/github/getting-started-with-github/associating-text-editors-with-git
EDITORS: Dict[str, List[str]] = {
    # -- $EDITOR and $VISUAL should be considered first --
    # Terminal based
    "sensible-editor": [],  # default editor in Debian-based systems like Ubuntu
    "nvim": [],  # if a person has nvim installed, chances are they like it as EDITOR
    "nano": [],  # beginner-friendly, vim users will likely have EDITOR=vim already set
    "vim": [],
    # GUI
    "subl": ["-w"],
    "code": ["--wait"],
    "mate": ["-w"],  # OS X specific
    "atom": ["--wait"],
    # Fallbacks (we tried reasonably hard to find a good editor...)
    "notepad": [],  # Windows
    "vi": [],  # POSIX
}
"""Programs to be tried (in sequence) when calling :obj:`edit` and :obj:`get_editor` in
the case the environment variables EDITOR and VISUAL are not set.
"""


class ShellCommand(object):
    """Shell command that can be called with flags like git('add', 'file')

    Args:
        command: command to handle
        shell: run the command in the shell (``True`` by default).
        cwd: current working dir to run the command

    The produced command can be called with the following keyword arguments:

        - **log** (*bool*): log activity when true. ``False`` by default.
        - **pretend** (*bool*): skip execution (but log) when pretending.
          ``False`` by default.

    The positional arguments are passed to the underlying shell command.
    In the case the path to the executable contains spaces of any other special shell
    character, ``command`` needs to be properly quoted.
    """

    def __init__(self, command: str, shell: bool = True, cwd: Optional[str] = None):
        self._command = command
        self._shell = shell
        self._cwd = cwd

    def run(self, *args, **kwargs) -> subprocess.CompletedProcess:
        """Execute command with the given arguments via :obj:`subprocess.run`."""
        command = f"{self._command} {join(args)}".strip()

        should_pretend = kwargs.pop("pretend", False)
        should_log = kwargs.pop("log", should_pretend)
        # ^ When pretending, automatically output logs
        #   (after all, this is the primary purpose of pretending)

        if should_log:
            logger.report("run", command, context=self._cwd)

        if should_pretend:
            return subprocess.CompletedProcess(command, 0, None, None)

        opts: dict = {
            "shell": self._shell,
            "cwd": self._cwd,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "universal_newlines": True,
            **kwargs,  # allow overwriting defaults
        }
        return subprocess.run(command, **opts)
        # ^ `check_output` does not seem to support terminal editors

    def __call__(self, *args, **kwargs) -> Iterator[str]:
        """Execute the command, returning an iterator for the resulting text output"""
        completed = self.run(*args, **kwargs)
        try:
            completed.check_returncode()
        except subprocess.CalledProcessError as ex:
            msg = "\n".join(e or "" for e in (completed.stdout, completed.stderr))
            raise ShellCommandException(msg) from ex

        return (line for line in (completed.stdout or "").splitlines())


def shell_command_error2exit_decorator(func: Callable):
    """Decorator to convert given ShellCommandException to an exit message

    This avoids displaying nasty stack traces to end-users
    """

    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except ShellCommandException as e:
            cause = e.__cause__
            reason = cause.output if cause and hasattr(cause, "output") else e
            print(f"{e}:\n{reason}")
            sys.exit(1)

    return func_wrapper


def get_git_cmd(**args):
    """Retrieve the git shell command depending on the current platform

    Args:
        **args: additional keyword arguments to :obj:`~.ShellCommand`
    """
    if sys.platform == "win32":  # pragma: no cover
        # ^  CI setup does not aggregate Windows coverage
        for cmd in ["git.cmd", "git.exe"]:
            git = ShellCommand(cmd, **args)
            try:
                git("--version")
            except ShellCommandException:
                continue
            return git
        else:
            return None
    else:
        git = ShellCommand("git", **args)
        try:
            git("--version")
        except ShellCommandException:
            return None
        return git


def command_exists(cmd: str) -> bool:
    """Check if command exists

    Args:
        cmd: executable name
    """
    if shutil.which(cmd) is None:
        return False
    else:
        return True


def get_executable(
    name: str, prefix: PathLike = sys.prefix, include_path=True
) -> Optional[str]:
    """Find an executable in the system, if available.

    Args:
        name: name of the executable
        prefix: look on this directory, exclusively or in additon to $PATH
            depending on the value of ``include_path``. Defaults to :obj:`sys.prefix`.
        include_path: when True the functions tries to look in the entire $PATH.


    Note:
        The return value might contain whitespaces. If this value is used in a shell
        environment, it needs to be quote properly to avoid the underlying shell
        interpreter splitting the executable path.
    """
    executable = shutil.which(name)
    if include_path and executable:
        return executable

    candidates = list(Path(prefix).resolve().glob(f"*/{name}*"))
    # ^  this works in virtual envs and both Windows and POSIX
    if candidates:
        path = (f.parent for f in sorted(candidates, key=lambda p: len(str(p))))
        return shutil.which(name, path=os.pathsep.join(str(p) for p in path))
        # ^  which will guarantee we find an executable and not only a regular file

    return None


def get_command(
    name: str, prefix: PathLike = sys.prefix, include_path=True, shell=True, **kwargs
) -> Optional[ShellCommand]:
    """Similar to :obj:`get_executable` but return an instance of :obj:`ShellCommand`
    if it is there to be found.
    Additional kwargs will be passed to the :obj:`ShellCommand` constructor.
    """
    executable = get_executable(name, prefix, include_path)
    if not executable:
        return None

    if shell:
        executable = join([executable])

    kwargs["shell"] = shell
    return ShellCommand(executable, **kwargs)


def get_editor(**kwargs):
    """Get an available text editor program"""
    from_env = os.getenv("VISUAL") or os.getenv("EDITOR")
    if from_env:
        return from_env  # user is responsible for proper quoting

    candidates = ((get_executable(e), opts) for e, opts in EDITORS.items())
    editor, opts = next((c for c in candidates if c[0]), (None, [""]))
    if editor:
        return join([editor, *opts])

    msg = "No text editor found in your system, please set EDITOR in your environment"
    raise ShellCommandException(msg)


def edit(file: PathLike, *args, **kwargs) -> Path:
    """Open a text editor and returns back a :obj:`Path` to file, after user editing."""
    editor = ShellCommand(get_editor())
    editor(file, *args, **{"stdout": None, "stderr": None, **kwargs})
    # ^  stdout/stderr=None => required for a terminal editor to open properly
    return Path(file)


def join(parts: Iterable[Union[str, PathLike]]) -> str:
    """Join different parts of a shell command into a string, quoting whitespaces."""
    if sys.platform == "win32":  # pragma: no cover
        # ^  CI setup does not aggregate Windows coverage
        return subprocess.list2cmdline(map(str, parts))

    return " ".join(shlex.quote(str(p)) for p in parts)
    # ^  TODO: Replace with `shlex.join(map(str, parts))` when Python >= 3.8


#: Command for git
git = get_git_cmd()



# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to use the teradata_rpo(S) and teradata_fake_result_sets
# escape functions to prepare a SQL request without executing it and obtain SQL statement metadata.
# This sample program assumes that StatementInfo parcel support is available from the Teradata Database.

import json
import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        cur.execute ('{fn teradata_rpo(S)}{fn teradata_fake_result_sets}select * from dbc.dbcinfo where infokey=?')
        row = cur.fetchone ()
        print ('SQL statement metadata from prepare operation:')
        print ()
        for i in range (0, len (cur.description)):
            print ('   Column [{}] {:>18} : {}'.format (i, cur.description [i][0], row [i]))
        print ()
        print ('Result set column metadata as pretty-printed JSON:')
        print (json.dumps (json.loads (row [7]), indent=4, sort_keys=True))
        print ()
        print ('Parameter marker metadata as pretty-printed JSON:')
        print (json.dumps (json.loads (row [8]), indent=4, sort_keys=True))



import sys

import snowflake.connector
from snowflake.sqlalchemy import URL
from sqlalchemy import create_engine
from snowfinch.logger import log

logger = log.get_logger()


def snowfinch_connections(dialect, conf):
    conn = None
    if conn is None:
        conn = ''
    if dialect == 'teradata':
        usr = conf['teradata']['user'].get()
        passwd = conf['teradata']['password'].get()
        host = conf['teradata']['host'].get()
        url: str = f'teradatasql://{host}/?user={usr}&password={passwd}&logmech' \
                   f'=LDAP&tmode=TERA&encryptdata=true'

        logger.info(url)

        try:
            logger.info("Connecting to Teradata database...")
            conn = create_engine(url.strip()).connect()
            logger.info("Successfully connected to Teradata.")
        except Exception as err:
            print("Failed to connect to Teradata DB: {}.".format(err))
            raise

    elif dialect == 'snowflake':
        usr = conf['snowflake']['user'].get()
        auth = conf['snowflake']['authenticator'].get()
        try:
            print("Connecting to SnowFlake database...")
            if usr.upper().startswith('SRC') or auth.upper() == 'LOCAL':
                logger.info("Running using Service Account")
                conn = create_engine(URL(
                    account=conf['snowflake']['account'].get(),
                    user=conf['snowflake']['user'].get(),
                    password=conf['snowflake']['password'].get(),
                    database=conf['snowflake']['database'].get(),
                    schema=conf['snowflake']['schema'].get(),
                    warehouse=conf['snowflake']['warehouse'].get(),
                    validate_default_parameters=True
                ))
            else:
                logger.info("Running using HA with external authenticator")
                conn = create_engine(URL(
                    account=conf['snowflake']['account'].get(),
                    user=conf['snowflake']['user'].get(),
                    password=conf['snowflake']['password'].get(),
                    database=conf['snowflake']['database'].get(),
                    schema=conf['snowflake']['schema'].get(),
                    warehouse=conf['snowflake']['warehouse'].get(),
                    authenticator="externalbrowser",
                    validate_default_parameters=True
                ))
                logger.info("Successfully connected to snowflake.")
        except snowflake.connector.errors.DatabaseError as err:
            logger.error("Failed to connect to Snowflake DB: {}.".format(err))
            raise

    elif dialect == 'mssql':
        import pyodbc
        # driver = 'ODBC+Driver+17+for+SQL+Server'
        host = conf['mssql']['host'].get()
        database = conf['mssql']['database'].get()
        usr = conf['mssql']['user'].get()
        passwd = conf['mssql']['password'].get()
        try:
            logger.info("Connecting to MSSQL Server database...")
            # conn = create_engine(conn_url)
            # conn = create_engine("mssql+pyodbc:///?odbc_connect=%s" % params)
            # conn = create_engine("mssql+pyodbc:///?odbc_connect={}".format(params))
            # conn = create_engine(conn_str)

            conn = pyodbc.connect('DRIVER={ODBC Driver 17 for SQL Server};'
                                  f'SERVER={host};'
                                  f'DATABASE={database};UID={usr};PWD={passwd}')
            logger.info("Successfully connected to MSSQL Server.")
        except Exception as err:
            logger.info("Failed to connect to MSSQL Server DB: {}.".format(err))
            sys.exit(1)
    return conn


def get_source_db_connections(dialect, config):
    # get the source  connection objects
    conn = None

    if conn is None:
        conn = ''

    if dialect == 'mssql':
        conn = snowfinch_connections(dialect, config)

    elif dialect == 'teradata':
        conn = snowfinch_connections(dialect, config)

    elif dialect == 'netezza':
        conn = snowfinch_connections(dialect, config)

    return conn



import re

# RegExes for SQL-Server dialect that Snowflake doesn't support

# NULL (explicit NULL constraint) -- ignore
null_constraint_re = re.compile('(.*)((?<!NOT)\s+NULL(?!::))(.*)', re.IGNORECASE)
is_null_condition_re = re.compile('.*IS NULL.*', re.IGNORECASE)

# NVARCHAR => VARCHAR
nvarchar_re = re.compile('(.*)\ (NVARCHAR)(.*)', re.IGNORECASE)

# NVARCHAR => VARCHAR
nchar_re = re.compile('(.*)\ (NCHAR)(.*)', re.IGNORECASE)

# ON PRIMARY => ignore
on_primary_re = re.compile('(.*)\ (ON PRIMARY)(.*)', re.IGNORECASE)

# DATETIME => TIMESTAMP
datetime_re = re.compile('(.*)\ (DATETIME)(.*)', re.IGNORECASE)

# BIT => BOOLEAN
bit_re = re.compile('(.*)\ (BIT)(.*)', re.IGNORECASE)


def parse_mssql(in_sql, out_sql, no_comments):
    # processing mode
    comment_lines = None
    term_re = None

    for line in in_sql:
        # state variables
        pre = None
        clause = None
        post = None
        comment = None

        sql = line.rstrip()
        sql = sql.replace('[', '').replace(']', '')

        # print >> sys.stdout, 'input: ' + sql

        if comment_lines:
            result = term_re.match(sql)
            if result:
                comment_lines = None
                term_re = None
            sql = '-- {0}'.format(sql)

        # NVARCHAR => VARCHAR
        result = nvarchar_re.match(sql)
        if result:
            pre = result.group(1)
            clause = result.group(2)
            post = result.group(3)
            sql = '{0} VARCHAR {1}\t\t-- {2}'.format(pre, post, clause)

        # NCHAR => CHAR
        result = nchar_re.match(sql)
        if result:
            pre = result.group(1)
            clause = result.group(2)
            post = result.group(3)
            sql = '{0} CHAR {1}\t\t-- {2}'.format(pre, post, clause)

        # DATETIME => TIMESTAMP
        result = datetime_re.match(sql)
        if result:
            pre = result.group(1)
            clause = result.group(2)
            post = result.group(3)
            sql = '{0} TIMESTAMP {1}\t\t-- {2}'.format(pre, post, clause)

        # BIT => BOOLEAN
        result = bit_re.match(sql)
        if result:
            pre = result.group(1)
            clause = result.group(2)
            post = result.group(3)
            sql = '{0} BOOLEAN {1}\t\t-- {2}'.format(pre, post, clause)

        # NULL (without NOT) => implicit nullable
        result = null_constraint_re.match(sql)
        if result and is_null_condition_re.match(sql):
            # we are in sql_query or DML, so not looking at a constraint
            result = None
        if result:
            pre = result.group(1)
            clause = result.group(2)
            post = result.group(3)
            sql = '{0}{1}\t\t-- {2}'.format(pre, post, clause)

        # ON PRIMARY => ignore
        result = on_primary_re.match(sql)
        if result:
            pre = result.group(1)
            clause = result.group(2)
            post = result.group(3)
            sql = '{0}{1}\t\t-- {2}'.format(pre, post, clause)

        # write out possibly modified line
        out_sql.write(sql)
        if comment:
            out_sql.write('\t\t-- {0}'.format(comment))
        out_sql.write('\n')
        continue


def append_comment(old_comment, new_comment, no_comments):
    if no_comments:
        return None
    if old_comment and new_comment:
        return '{0} // {1}'.format(old_comment, new_comment)
    if not old_comment:
        return new_comment
    return old_comment


# if __name__ == "__main__":
#     parse_mssql(inputfile, outfile, no_comments)
#     inputfile.close()
#     outfile.close()
#     print("done parsing " + inputfile.name)



# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to obtain the version numbers for the
# Teradata SQL Driver for Python and the Teradata Database.

import teradatasql

with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
    with con.cursor () as cur:
        cur.execute ('{fn teradata_nativesql}Driver version {fn teradata_driver_version}  Database version {fn teradata_database_version}')
        print (cur.fetchone () [0])



# Copyright 2018 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to use the Python csv module to parse data values
# from a CSV file, and then insert those data values into a database table.
# This sample program requires the airports.csv file to be located in the current directory.

import csv
import teradatasql

with open ('airports.csv', newline='') as f:
    with teradatasql.connect ('{"host":"whomooz","user":"guest","password":"please"}') as con:
        with con.cursor () as cur:
            cur.execute ("create volatile table Airports (City varchar(100), Airport varchar(100), AirportCode varchar(10)) on commit preserve rows")
            cur.execute ("insert into Airports (?, ?, ?)", [ row for row in csv.reader (f) ])
            cur.execute ("select AirportCode, Airport, City from Airports order by AirportCode")
            [ print (row) for row in cur.fetchall () ]



# Copyright 2020 by Teradata Corporation. All rights reserved.

# This sample program demonstrates how to insert and retrieve XML values.
# Use the escape function teradata_parameter to override the data type for a parameter marker bind value.

import json
import teradatasql

with teradatasql.connect (host="whomooz", user="guest", password="please") as con:
    with con.cursor () as cur:
        cur.execute ("create volatile table voltab (c1 integer, c2 xml) on commit preserve rows")

        cur.execute ("{fn teradata_parameter(2,XML)}insert into voltab values (?, ?)", [
            [ 1, "<hello>world</hello>" ],
            [ 2, "<hello>moon</hello>" ],
        ])

        cur.execute ("{fn teradata_fake_result_sets}select * from voltab order by 1")

        # obtain column metadata from the fake result set
        asTypeNames = [ m ["TypeName"] for m in json.loads (cur.fetchone () [7]) ]

        cur.nextset () # advance to the real result set

        for nRow, row in enumerate (cur.fetchall (), 1):
            for iColumn, oValue in enumerate (row):
                print ("Row {} Column {} {:<7} {}".format (nRow, cur.description [iColumn][0], asTypeNames [iColumn], oValue))



"""
Command-Line-Interface(CLI) of SnowFinch
"""
import argparse
import confuse
from typing import List, Optional

from snowfinch import snowfinch_version
from snowfinch.parsers.parser import *
from snowfinch.aws.s3upload import *
from snowfinch.exceptions import *
from snowfinch.utils import regex
from snowfinch.shell import shell_command_error2exit_decorator


USAGE = (
    "\nTo see help text, you can run:\n"
    "  snowfinch help\n"
    "  snowfinch <command> help\n"
    "  snowfinch <command> <subcommand> help\n"
)


def parse_args(args):
    """Parse command line parameters

    Args:
      args (List[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        usage=USAGE,
        description="Command Line Environment for SnowFinch.")

    # Required arguments
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')

    # Optional  argument
    optional.add_argument(
        "-v",
        "--version",
        action='version',
        version="snowfinch {ver}".format(ver=snowfinch_version),
        help="use this option to get SnowFinch latest version and exit",
    )
    required.add_argument(
        "-c",
        "--config",
        type=str,
        metavar='',
        required=True,
        help='use this option to read SnowFinch arguments from a given config file like .yaml, .yml, .ini, .toml',
    )

    optional.add_argument(
        '-s',
        '--secret',
        type=str,
        metavar='',
        help="use this option to handle sensitive configuration like DB Credentials, Secrete key etc")

    optional.add_argument(
        '-p',
        '--profile',
        type=str,
        metavar='',
        required=False,
        help='use this option to pick a specific environment configuration like dev, sit and prod',
    )

    required.add_argument(
        '-m',
        '--module',
        type=str,
        metavar='',
        required=True,
        help="use this option to run a specific SnowFinch module like bteq, ddl, s3 or all",
    )

    optional.add_argument(
        "-dt",
        "--drop-table",
        type=bool,
        default=False,
        metavar='',
        help='use this option if you want SnowFinch to drop and create tables',
    )

    required.add_argument(
        '-d',
        '--dialect',
        type=str,
        metavar='',
        required=True,
        help='use this option to select to select the dialect like mssql, teradata',
    )

    optional.add_argument(
        "-ll",
        "--log-level",
        type=str,
        metavar='',
        default="DEBUG",
        help="use this option to set the SnowFinch logging level to the desired verbosity",
    )

    optional.add_argument(
        '-cc',
        '--concurrency',
        metavar='',
        type=int,
        default=1,
        help="use this option to set the number of concurrent threads to use",
    )

    optional.add_argument(
        '-e',
        '--exclude',
        action='append',
        type=regex,
        metavar='',
        help="use this option to exclude a list of files using a regex method", )

    optional.add_argument(
        '-i',
        '--include',
        action='append',
        type=regex,
        metavar='',
        help="use this option to include list of files using a regex method",
    )

    parser.print_help()
    parser.print_usage()
    return parser.parse_args(args)


# def setup_logging(loglevel):
#     """Setup basic logger
#     Args:
#       loglevel (int): minimum loglevel for emitting messages
#     """
#     logformat = "%(asctime)s %(levelname)s:[%(name)s]:%(message)s"
#     logging.basicConfig(
#         level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
#     )


def main(args: List[str]):
    """
    Main entry point for external applications
        Args:
            args: command line arguments
    """
    args = parse_args(args)

    if args.concurrency < 1:
        logger.info("Concurrency must be positive.", file=sys.stderr)

    try:
        if args.log_level.upper() == "WARN":
            # logging.basicConfig(level=logging.WARN)
            logger.setLevel(level=logging.WARN)

        elif args.log_level.upper() == "INFO":
            # logging.basicConfig(level=logging.INFO)
            logger.setLevel(level=logging.INFO)
        elif args.log_level.upper()  == "DEBUG":
            # logging.basicConfig(level=logging.DEBUG)
            logger.setLevel(level=logging.DEBUG)

        logger.info("Log level set: {}".format(logger.getEffectiveLevel()))
    except ValueError:
        logger.error("Invalid log level: {}".format(args.verbose))
        sys.exit(1)

    # initiate the SnowFinch
    logger.info("Starting the SnowFinch Application...\n")

    # get the configs file
    app_config = args.config
    # profile = args.profile
    dialect = args.dialect

    if not os.path.isfile(app_config):
        print("File path {} does not exist. Exiting...".format(app_config))
        sys.exit()

    logger.info(f"Fetching application configs...")
    logger.info(f"application configs file: {app_config}")

    # initiate the configuration
    config = confuse.Configuration('SnowFinch', __name__)

    # Add conf items from specified file
    config.set_file(app_config)

    if args.module == 'bteq':
        logger.info("Running SnowFinch in bteq-only mode...\n")
        get_bteq_build(config)

    elif args.module == 'sproc':
        logger.info("Running SnowFinch in sproc-only mode...\n")
        get_sproc_build(dialect, config)

    elif args.module == 'ddl':
        logger.info("Running SnowFinch in ddl-only mode...\n")
        get_ddl_build(dialect, config, drop_table=args.drop_table)

    elif args.module == 's3':
        logger.info("Running SnowFinch in s3-only mode...\n")
        get_s3_upload(config)

    else:
        logger.info("Running SnowFinch full mode...\n")
        get_full_build(dialect, config, drop_table=args.drop_table)

    logger.info("SnowFinch executed successfully")


@shell_command_error2exit_decorator
@exceptions2exit([RuntimeError])
def run(args: Optional[List[str]] = None):
    """Entry point for console script"""
    print(f'Snowfinch Version: {snowfinch_version}')
    main(args or sys.argv[1:])


if __name__ == "__main__":
    main(sys.argv[1:])
